# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pytest

from surfactant.utils.regex import (
    extract_fixed_literals,
    extract_fixed_prefixes,
    handle_escaped_literal,
)


class TestHandleEscapedLiteral:
    def test_basic_escaped_literals(self):
        assert handle_escaped_literal(r"\.", 0, 2) == (".", 2)
        assert handle_escaped_literal(r"\-", 0, 2) == ("-", 2)

    def test_special_escape_sequences(self):
        assert handle_escaped_literal(r"\a", 0, 2) == (None, 0)
        assert handle_escaped_literal(r"\n", 0, 2) == ("\n", 2)
        assert handle_escaped_literal(r"\r", 0, 2) == ("\r", 2)
        assert handle_escaped_literal(r"\t", 0, 2) == ("\t", 2)
        assert handle_escaped_literal(r"\v", 0, 2) == ("\v", 2)
        assert handle_escaped_literal(r"\f", 0, 2) == ("\f", 2)
        assert handle_escaped_literal(r"\0", 0, 2) == (None, 0)

    def test_control_characters(self):
        # Test redundant check issue where next_char is checked twice
        assert handle_escaped_literal(r"\cA", 0, 3) == (chr(ord("A") - ord("A") + 1), 3)
        assert handle_escaped_literal(r"\cZ", 0, 3) == (chr(ord("Z") - ord("A") + 1), 3)

        # Test lowercase control character (should return None, 0)
        assert handle_escaped_literal(r"\ca", 0, 3) == (None, 0)

    def test_hex_escapes(self):
        assert handle_escaped_literal(r"\x41", 0, 4) == ("A", 4)  # ASCII 65 = 'A'
        assert handle_escaped_literal(r"\x7A", 0, 4) == ("z", 4)  # ASCII 122 = 'z'

        # Test invalid hex sequence
        assert handle_escaped_literal(r"\xZZ", 0, 4) == (None, 0)

    def test_regex_special_chars(self):
        # These should all return (None, 0) as they're special regex entities
        assert handle_escaped_literal(r"\d", 0, 2) == (None, 0)
        assert handle_escaped_literal(r"\w", 0, 2) == (None, 0)
        assert handle_escaped_literal(r"\s", 0, 2) == (None, 0)
        assert handle_escaped_literal(r"\b", 0, 2) == (None, 0)

    def test_overlapping_character_sets(self):
        # Test "0" which appears in multiple conditions
        assert handle_escaped_literal(r"\0", 0, 2) == (
            None,
            0,
        )  # null char (potentially also backref though)

        # Test digit reference (should return None, 0)
        assert handle_escaped_literal(r"\1", 0, 2) == (None, 0)  # backreference

    def test_unicode(self):
        # Testing the incomplete Unicode support
        assert handle_escaped_literal(r"\u0041", 0, 6) == (
            None,
            0,
        )  # Should return "A" once implemented properly

        # Test curly brace Unicode format
        assert handle_escaped_literal(r"\u{0041}", 0, 8) == (None, 0)  # Not properly handled


class TestExtractFixedPrefixes:
    def test_basic_strings(self):
        assert extract_fixed_prefixes("abc") == ["abc"]
        assert extract_fixed_prefixes("Hello") == ["Hello"]
        assert extract_fixed_prefixes("abc123") == ["abc123"]

    def test_with_caret(self):
        assert extract_fixed_prefixes("^abc") == ["abc"]
        assert extract_fixed_prefixes("^Hello") == ["Hello"]

    def test_with_escaped_characters(self):
        assert extract_fixed_prefixes(r"abc\n") == ["abc\n"]
        assert extract_fixed_prefixes(r"\tHello") == ["\tHello"]
        assert extract_fixed_prefixes(r"\x41BC") == ["ABC"]

    def test_character_classes(self):
        result = extract_fixed_prefixes("[ab]cd")
        assert len(result) == 2
        assert "acd" in result and "bcd" in result

        result = extract_fixed_prefixes("[a-c]de")
        assert len(result) == 3
        assert "ade" in result and "bde" in result and "cde" in result

        # Test with negated class (should return empty list)
        assert not extract_fixed_prefixes("[^abc]")

    def test_character_class_with_escaped_chars(self):
        result = extract_fixed_prefixes(r"[\t\n]abc")
        assert len(result) == 2
        assert "\tabc" in result and "\nabc" in result

        # Test escaped literals within character class
        result = extract_fixed_prefixes(r"[\[\]]abc")
        assert len(result) == 2
        assert "[abc" in result and "]abc" in result

    def test_character_ranges(self):
        # Test with hyphen at different positions
        result = extract_fixed_prefixes("[-ab]")  # Literal hyphen at start
        assert len(result) == 3
        assert "-" in result and "a" in result and "b" in result

        result = extract_fixed_prefixes("[ab-]")  # Literal hyphen at end
        assert len(result) == 3
        assert "a" in result and "b" in result and "-" in result

    def test_capture_groups(self):
        assert extract_fixed_prefixes("(abc)") == ["abc"]

        # Test alternation in groups
        result = extract_fixed_prefixes("(abc|def)")
        assert len(result) == 2
        assert "abc" in result and "def" in result

        # Test non-alphanumeric content in groups (restricted handling)
        assert extract_fixed_prefixes("(abc|d!e)") == ["abc"]

        # Test nested groups (should return empty list)
        assert not extract_fixed_prefixes("(a(bc))")

        # Test non-capturing groups (should return empty list)
        assert not extract_fixed_prefixes("(?:abc)")

    def test_limited_prefix_extraction(self):
        # Test characters that are accepted (alphanumeric, underscore, hyphen)
        assert extract_fixed_prefixes("abc_def") == ["abc_def"]
        assert extract_fixed_prefixes("abc-def") == ["abc-def"]

        # Test characters that break prefix extraction
        assert extract_fixed_prefixes("abc.def") == ["abc"]
        assert extract_fixed_prefixes("abc!def") == ["abc"]

    def test_complex_patterns(self):
        result = extract_fixed_prefixes("^[aA]bc")
        assert len(result) == 2
        assert "abc" in result and "Abc" in result

        result = extract_fixed_prefixes("^(abc|def)123")
        assert len(result) == 2
        assert "abc123" in result and "def123" in result

    def test_special_regex_features(self):
        # Test with quantifiers
        assert extract_fixed_prefixes("abc+") == ["abc"]
        assert extract_fixed_prefixes("abc*") == ["abc"]

        # Test with special regex characters
        assert extract_fixed_prefixes("abc\\d") == ["abc"]
        assert not extract_fixed_prefixes("\\w+abc")

    def test_bounds_checking(self):
        # Empty pattern
        assert not extract_fixed_prefixes("")

        # Only special characters
        assert not extract_fixed_prefixes("\\d\\w\\s")

        # Invalid regex
        with pytest.raises(ValueError):
            extract_fixed_prefixes("[unclosed")

        with pytest.raises(ValueError):
            extract_fixed_prefixes("(unclosed")


# Test case that runs through a few cases, some of which may already be covered above
def test_extract_fixed_prefixes():
    # Test case 1: Regex with capture group and multiple prefixes
    regex1 = r"^(abc|def)\d+world"
    expected1 = ["abc", "def"]
    assert extract_fixed_prefixes(regex1) == expected1

    # Test case 2: Regex with escaped character
    regex2 = r"hello\.\d+world"
    expected2 = ["hello."]
    assert extract_fixed_prefixes(regex2) == expected2

    # Test case 3: Regex with character class
    regex3 = r"[xyz]123"
    expected3 = ["x123", "y123", "z123"]
    assert extract_fixed_prefixes(regex3) == expected3

    # Test case 4: Regex with no fixed prefix
    regex4 = r"\d+world"
    expected4 = []
    assert extract_fixed_prefixes(regex4) == expected4

    # Test case 5: Regex with starting '^'
    regex5 = r"^start123"
    expected5 = ["start123"]
    assert extract_fixed_prefixes(regex5) == expected5

    # Test case 6: Invalid regex pattern
    regex6 = r"^(abc"
    with pytest.raises(ValueError):
        extract_fixed_prefixes(regex6)

    # Test case 7: Regex with alphanumeric characters and special characters
    regex7 = r"prefix_123-abc"
    expected7 = ["prefix_123-abc"]
    assert extract_fixed_prefixes(regex7) == expected7

    # Test case 8: Regex with capture group followed by literal
    regex8 = r"^(foo|bar)baz"
    expected8 = ["foobaz", "barbaz"]
    assert extract_fixed_prefixes(regex8) == expected8

    # Test case 9: Regex with nested capture groups
    regex9 = r"^((foo|bar)baz)"
    expected9 = []
    assert extract_fixed_prefixes(regex9) == expected9

    # Test case 10: Regex with multiple capture groups
    regex10 = r"^(foo|bar)xyz(baz|qux)"
    expected10 = ["fooxyz", "barxyz"]
    assert extract_fixed_prefixes(regex10) == expected10

    # Test case 11: Regex with escaped backslash
    regex11 = r"prefix\\d+suffix"  # raw string so no \\ => \
    expected11 = ["prefix\\d"]  # regular string, so \\ => \
    assert extract_fixed_prefixes(regex11) == expected11


# Tests for the tree-based implementation of literal extraction, using Python's re parser below this point
def assert_order_independent(actual, expected):
    """
    Helper function to assert that two lists are equal, ignoring order.
    Also asserts that the second part of the tuple matches exactly.
    """
    assert set(actual[0]) == set(expected[0]), f"Expected {expected[0]}, but got {actual[0]}"
    assert actual[1] == expected[1], f"Expected {expected[1]}, but got {actual[1]}"


def test_extract_fixed_literals():
    # Test case 1: Regex with fixed literals
    regex1 = r"hello123world"
    expected1 = (["hello123world"], True)
    assert extract_fixed_literals(regex1) == expected1

    # Test case 2: Regex with alternation
    regex2 = r"^(foo|bar)baz"
    expected2 = (["foobaz", "barbaz"], True)
    assert_order_independent(extract_fixed_literals(regex2), expected2)

    # Test case 3: Regex with character class
    regex3 = r"[abc]123"
    expected3 = (["a123", "b123", "c123"], True)
    assert_order_independent(extract_fixed_literals(regex3), expected3)

    # Test case 4: Regex with no fixed literals
    regex4 = r"\d+world"
    expected4 = (["world"], False)
    assert extract_fixed_literals(regex4) == expected4

    # Test case 5: Regex with repetition
    regex5 = r"(abc){2,3}"
    expected5 = (["abcabc", "abcabcabc"], True)
    assert_order_independent(extract_fixed_literals(regex5), expected5)

    # Test case 6: Regex with nested capture groups
    regex6 = r"^((foo|bar)baz)"
    expected6 = (["foobaz", "barbaz"], True)
    assert_order_independent(extract_fixed_literals(regex6), expected6)

    # Test case 7: Regex with escaped characters
    regex7 = r"hello\.\d+world"
    expected7 = (["hello."], True)
    assert extract_fixed_literals(regex7) == expected7

    # Test case 8: Regex with anchors
    regex8 = r"^start123$"
    expected8 = (["start123"], True)
    assert extract_fixed_literals(regex8) == expected8

    # Test case 9: Regex with alternation and literals
    regex9 = r"(foo|bar)(baz|qux)"
    expected9 = (["foobaz", "fooqux", "barbaz", "barqux"], True)
    assert_order_independent(extract_fixed_literals(regex9), expected9)

    # Test case 10: Regex with invalid pattern
    regex10 = r"^(abc"
    expected10 = ([], False)
    assert extract_fixed_literals(regex10) == expected10

    # Test case 11: Regex with optional group
    regex11 = r"(foo)?bar"
    expected11 = (["bar", "foobar"], True)
    assert_order_independent(extract_fixed_literals(regex11), expected11)

    # Test case 12: Regex with range in character class
    regex12 = r"[a-c]123"
    expected12 = (["a123", "b123", "c123"], True)
    assert_order_independent(extract_fixed_literals(regex12), expected12)

    # Test case 13: Regex with repetition and fixed literals
    regex13 = r"(abc){2}"
    expected13 = (["abcabc"], True)
    assert extract_fixed_literals(regex13) == expected13

    # Test case 14: Regex with escaped backslash
    regex14 = r"prefix\\suffix"
    expected14 = (["prefix\\suffix"], True)
    assert extract_fixed_literals(regex14) == expected14

    # Test case 15: Regex with multiple fixed literals
    regex15 = r"hello123|world456"
    expected15 = (["hello123", "world456"], True)
    assert_order_independent(extract_fixed_literals(regex15), expected15)

    # Test case 16: Regex with comprehensive branching
    regex16 = r"he(l|[0-5])o"
    expected16 = (["helo", "he0o", "he1o", "he2o", "he3o", "he4o", "he5o"], True)
    assert_order_independent(extract_fixed_literals(regex16), expected16)

    # Test case 17: Regex with non-comprehensive branching
    regex17 = r"he(l|[0-5]+)o"
    expected17 = (["hel", "he0", "he1", "he2", "he3", "he4", "he5"], True)
    assert_order_independent(extract_fixed_literals(regex17), expected17)


def test_extract_fixed_literals_actual_cases():
    # Test case: rflow (only possible prefixes exceed max possibilities threshold of 10)
    rflow = "[0-9]\\.[0-9]+\\ Copyright\\ by\\ Nikki\\ Chumakov"
    expected_rflow = ([" Copyright by Nikki Chumakov"], False)
    assert extract_fixed_literals(rflow) == expected_rflow

    # Test case: miniupnpd
    miniupnpd = r"SERVER:.*UPnP\/[0-9](\.[0-9]+)+?\ MiniUPnPd\/[0-9](\.[0-9]+)+?"
    expected_miniupnpd = (["SERVER:"], True)
    assert extract_fixed_literals(miniupnpd) == expected_miniupnpd

    # Test case: nlohmann
    # Only guaranteed literals are "nlohmann[0-9]" due to infinite max repetitions of the numbers
    # may be worth checking to see if upstream pattern is correct and not supposed to be "nlohman::"
    nlohmann = r"nlohmann[0-9]+json_abi_v[1-3]+_[0-9]+(_[0-9])?"
    expected_nlohmann = (
        [f"nlohmann{digit}" for digit in range(10)],
        True,
    )
    assert_order_independent(extract_fixed_literals(nlohmann), expected_nlohmann)

    # Test case: mkenv (an optional component)
    mkenv = "mk(env)?image\\ version\\ 20[0-9]+\\.[0-9]+"
    expected_mkenv = (["mkimage version 20", "mkenvimage version 20"], True)
    assert_order_independent(extract_fixed_literals(mkenv), expected_mkenv)

    # Test case: version with the dot in the capture group
    dot_in_capture_group = "gcc version\\ [0-9](\\.[0-9]+)+?"
    expected_dot_in_capture_group = (
        [f"gcc version {digit}" for digit in range(10)],
        True,
    )
    assert_order_independent(
        extract_fixed_literals(dot_in_capture_group), expected_dot_in_capture_group
    )

    # Test case: version with a dot outside the capture group
    dot_outside_capture_group = "gcc version\\ [0-9]\\.([0-9]+)"
    expected_dot_outside_capture_group = (
        [f"gcc version {digit}." for digit in range(10)],
        True,
    )
    assert_order_independent(
        extract_fixed_literals(dot_outside_capture_group), expected_dot_outside_capture_group
    )

    # Test case: chrony
    chrony = "chrony[cd]\\ \\(chrony\\)\\ version\\ [0-9]\\.[0-9]+"
    expected_chrony = (["chronyc (chrony) version ", "chronyd (chrony) version "], True)
    assert_order_independent(extract_fixed_literals(chrony), expected_chrony)



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import re
from typing import Any, Dict, List, Optional

# from pluggy import get_plugin_manager, get_plugin
from loguru import logger

import surfactant.plugin
from surfactant.database_manager.database_utils import BaseDatabaseManager, DatabaseConfig
from surfactant.sbomtypes import SBOM, Software

# Global configuration
DATABASE_URL_RETIRE_JS = "https://raw.githubusercontent.com/RetireJS/retire.js/master/repository/jsrepository-master.json"
JS_DB_DIR = "js_library_patterns"  # The directory name to store the database toml file and database json files for this module


@surfactant.plugin.hookimpl
def short_name() -> str:
    return "js_file"


class RetireJSDatabaseManager(BaseDatabaseManager):
    """Manages the retirejs library database."""

    def __init__(self):
        name = (
            short_name()
        )  # Set to '__name__' (without quotation marks), if short_name is not implemented

        config = DatabaseConfig(
            database_dir=JS_DB_DIR,  # The directory name to store the database toml file and database json files for this module.
            database_key="retirejs",  # The key for this classes database in the version_info toml file.
            database_file="retirejs_db.json",  # The json file name for the database.
            source=DATABASE_URL_RETIRE_JS,  # The source of the database (put "file" or the source url)
            plugin_name=name,
        )

        super().__init__(config)

    def parse_raw_data(self, raw_data: str) -> Dict[str, Dict[str, List[str]]]:
        """
        Parses a RetireJS JSON dump into a nested dict:
        { library_name: { 'filename': [regex...], 'filecontent': [...], 'hashes': [...] } }
        Invalid JSON or invalid regex entries are logged and skipped.
        """

        try:
            db = json.loads(raw_data)
        except json.JSONDecodeError as err:
            logger.error(f"Failed to parse downloaded database JSON: {err}")

        if not isinstance(db, dict):
            logger.error("Expected top-level JSON object for RetireJS data")
            return {}

        VERSION_PLACEHOLDER = "\u00a7\u00a7version\u00a7\u00a7"
        VERSION_NUMBER_PATTERN = r"\d+(?:\.\d+)*"

        clean_db: Dict[str, Dict[str, List[str]]] = {}

        for library, lib_entry in db.items():
            if "extractors" in lib_entry:
                clean_db[library] = {}
                for entry in ["filename", "filecontent", "hashes"]:
                    if entry in lib_entry["extractors"]:
                        # Initialize the list in clean_db
                        clean_db[library][entry] = []

                        # Iterate through each pattern, do the replacement, and append
                        for reg in lib_entry["extractors"][entry]:
                            replaced = reg.replace(VERSION_PLACEHOLDER, VERSION_NUMBER_PATTERN)

                            try:
                                re.compile(replaced.encode("utf-8"))  # Validate regex
                                clean_db[library][entry].append(replaced)
                            except re.error as rex:
                                logger.error(
                                    "Invalid regex for %s[%s]: %r — %s",
                                    library,
                                    entry,
                                    replaced,
                                    rex,
                                )

        return clean_db


js_db_manager = RetireJSDatabaseManager()


def supports_file(filetype: List[str]) -> bool:
    return "JAVASCRIPT" in filetype


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: List[str]) -> object:
    if not supports_file(filetype):
        return None
    return extract_js_info(filename)


def extract_js_info(filename: str) -> object:
    js_info: Dict[str, Any] = {"jsLibraries": []}
    js_lib_database = js_db_manager.get_database()

    if js_lib_database is None:
        return None

    # Try to match file name
    libs = match_by_attribute("filename", filename, js_lib_database)
    if len(libs) > 0:
        js_info["jsLibraries"] = libs
        return js_info

    # Try to match file contents
    try:
        with open(filename, "r") as js_file:
            filecontent = js_file.read()
        libs = match_by_attribute("filecontent", filecontent, js_lib_database)
        js_info["jsLibraries"] = libs
    except FileNotFoundError:
        logger.warning("File not found: %s", filename)
    except UnicodeDecodeError:
        logger.warning("File does not appear to be UTF-8: %s", filename)
    return js_info


def match_by_attribute(attribute: str, content: str, database: Dict) -> List[Dict]:
    libs = []
    for name, library in database.items():
        if attribute in library:
            for pattern in library[attribute]:
                matches = re.search(pattern, content)
                if matches:
                    if len(matches.groups()) > 0:
                        libs.append({"library": name, "version": matches.group(1)})
                        # skip remaining patterns, move on to the next library
                        break
    return libs


def strip_irrelevant_data(retirejs_db: dict) -> dict:
    clean_db = {}
    reg_temp = "\u00a7\u00a7version\u00a7\u00a7"
    version_regex = r"\d+(?:\.\d+)*"
    for library, lib_entry in retirejs_db.items():
        if "extractors" in lib_entry:
            clean_db[library] = {}
            patterns = lib_entry["extractors"]
            possible_entries = [
                "filename",
                "filecontent",
                "hashes",
            ]
            for entry in possible_entries:
                if entry in patterns:
                    entry_list = []
                    for reg in patterns[entry]:
                        entry_list.append(reg.replace(reg_temp, version_regex))
                    clean_db[library][entry] = entry_list
    return clean_db


@surfactant.plugin.hookimpl
def update_db() -> str:
    return js_db_manager.download_and_update_database()


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None) -> None:
    js_db_manager.initialize_database(command_name)



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import urlparse

from loguru import logger

from surfactant.configmanager import ConfigManager
from surfactant.database_manager.utils import (
    calculate_hash,
    download_content,
    get_source_for,
    load_db_version_metadata,
    save_db_version_metadata,
)


@dataclass
class DatabaseConfig:
    """
    Represents the configuration for a database used by a plugin.

    Attributes:
        database_dir (str): The database directory name for the class of databases (e.g., "js_library", "native_library").
        database_key (str): The key identifying the database (e.g., "retirejs").
        database_file (str): The file name of the database (e.g., "js_library_patterns_retirejs.json").
        source (str): The source URL of the database, or "file" if it's a local file.
        plugin_name (Optional[str]): The canonical name or short name of the plugin handling the database.
    """

    database_dir: str
    database_key: str
    database_file: str
    source: str
    plugin_name: Optional[str] = None

    def __post_init__(self) -> None:
        # Validate that source is either a URL or "file"
        if self.source != "file":
            parsed_url = urlparse(self.source)
            # Check that the scheme is valid (http or https)
            if parsed_url.scheme not in {"http", "https"}:
                raise ValueError(
                    f"Invalid URL scheme: {parsed_url.scheme}. Expected 'http' or 'https'."
                )
            # Check that netloc is present
            if not parsed_url.netloc:
                raise ValueError(
                    f"Invalid URL for source: {self.source}",
                )

        # Ensure database_file ends with .json
        if not self.database_file.endswith(".json"):
            raise ValueError(
                f"database_file '{self.database_file}' must end with '.json'.",
            )


class BaseDatabaseManager(ABC):
    """Abstract base class for managing databases."""

    def __init__(self, config: DatabaseConfig) -> None:
        self.config = config

        # Attempt to retrieve an override URL using the database_dir (e.g., "js_library_patterns")
        # and the database_key (e.g., "retirejs").
        override_url = get_source_for(self.config.database_dir, self.config.database_key)
        if override_url:
            self.config.source = override_url
            logger.debug(
                "Using external URL override for {}: {}", self.config.database_key, override_url
            )
        else:
            logger.debug("Using hard-coded URL for {}", self.config.database_key)

        self.new_hash: Optional[str] = None
        self.download_timestamp: Optional[str] = None
        self._database: Optional[Dict[str, Any]] = None

        # Ensure the parent directory exists
        path = self.database_version_file_path
        path.parent.mkdir(parents=True, exist_ok=True)

    @property
    def data_dir(self) -> Path:
        """Returns the base directory for storing database (.json) and version tracking (TOML) files."""
        return ConfigManager().get_data_dir_path() / "databases"

    @property
    def database_version_file_path(self) -> Path:
        """Path to the database version file (e.g., TOML file)."""
        return self.data_dir / f"{self.config.database_dir}" / "version_info.toml"

    @property
    def database_file_path(self) -> Path:
        """Path to the JSON database file."""
        return self.data_dir / f"{self.config.database_dir}" / f"{self.config.database_file}"

    @property
    def database_info(self) -> Dict[str, Any]:
        """Returns metadata about the database patterns."""
        return {
            "database_key": self.config.database_key,
            "database_file": self.config.database_file,
            "source": self.config.source,
            "hash_value": self.new_hash,
            "timestamp": self.download_timestamp,
        }

    def initialize_database(self, command_name: Optional[str] = None) -> None:
        """
        Initialization hook to load the pattern database.

        Args:
            command_name (Optional[str], optional): The name of the command invoking the initialization.
                If set to "update-db", the database will not be loaded.

        Returns:
            None
        """
        if command_name != "update-db":
            logger.debug("Initializing {}...", self.config.plugin_name)
            self.load_db()
            logger.debug("Initializing {} complete.", self.config.plugin_name)

    def load_db(self) -> None:
        """Loads the database from a JSON file."""
        try:
            with self.database_file_path.open("r") as db_file:
                self._database = json.load(db_file)
        except FileNotFoundError:
            logger.warning(
                "{} database could not be loaded. Run `surfactant plugin update-db {}` to fetch the database.",
                self.config.database_key,
                self.config.plugin_name,
            )
            self._database = None

    def get_database(self) -> Optional[Dict[str, Any]]:
        """Returns the loaded database."""
        if self._database is None:
            self.load_db()
        return self._database

    def save_database(self, data: Dict[str, Any]) -> None:
        """Saves the database to a JSON file."""
        self.data_dir.mkdir(parents=True, exist_ok=True)
        with self.database_file_path.open("w") as db_file:
            json.dump(data, db_file, indent=4)
        logger.info("{} database saved successfully.", self.config.database_key)

    @abstractmethod
    def parse_raw_data(self, raw_data: str) -> Dict[str, Any]:
        """Parses raw database data into a structured format."""
        # No implementation needed for abstract methods.

    def download_and_update_database(self) -> str:
        raw_data = download_content(self.config.source)
        if not raw_data:
            return "No update occurred. Failed to download database."

        new_hash = calculate_hash(raw_data)
        current_data = load_db_version_metadata(
            self.database_version_file_path, self.config.database_key
        )

        if current_data and new_hash == current_data.get("hash"):
            return "No update occurred. Database is up-to-date."

        parsed_data = self.parse_raw_data(raw_data)
        if parsed_data is None:
            return "No update occurred. Failed to parse raw data."

        self.save_database(parsed_data)
        self.new_hash = new_hash
        self.download_timestamp = datetime.now(timezone.utc).isoformat()
        save_db_version_metadata(self.database_version_file_path, self.database_info)
        return "Update complete."



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, List, Union

from surfactant.utils.regex import extract_fixed_literals


# pylint: disable=too-few-public-methods
class AhoCorasickNode:
    """Node in the Aho-Corasick trie structure."""

    def __init__(self):
        self.goto = {}  # Goto function
        self.out = []  # Output function - list of pattern indices at this node
        self.fail = None  # Failure function
        self.depth = 0  # Depth of the node in the trie


class AhoCorasick:
    """Aho-Corasick automaton for fast multiple string matching."""

    def __init__(self, is_bytes: bool = False, encoding: str = "utf-8"):
        self.root = AhoCorasickNode()
        self.built = False
        self.is_bytes = is_bytes  # Flag to indicate if this automaton works with bytes
        self.pattern_prefixes = {}  # Maps pattern_id -> prefix used
        self.encoding = encoding  # Encoding to use for string/bytes conversion

    def add_pattern(self, pattern: Union[str, bytes], pattern_id: Any, prefix: str) -> None:
        """
        Add a pattern to the trie.

        Args:
            pattern: A string or bytes pattern to add
            pattern_id: An identifier for this pattern
            prefix: The prefix string used for this pattern
        """
        # Ensure pattern is in the correct type (str or bytes)
        if self.is_bytes and isinstance(pattern, str):
            pattern = pattern.encode(self.encoding)
        elif not self.is_bytes and isinstance(pattern, bytes):
            pattern = pattern.decode(self.encoding, errors="ignore")

        node = self.root
        for i, char in enumerate(pattern):
            # For bytes, comparison char will be int values for the byte
            if char not in node.goto:
                node.goto[char] = AhoCorasickNode()
                node.goto[char].depth = i + 1
            node = node.goto[char]
        node.out.append(pattern_id)

        # Store the prefix used for this pattern_id
        self.pattern_prefixes[pattern_id] = prefix

    def build_automaton(self) -> None:
        """Build the Aho-Corasick automaton by computing failure functions."""
        queue = []
        # Set failure of all depth 1 nodes to root
        for _char, node in self.root.goto.items():
            node.fail = self.root
            queue.append(node)

        # Build failure function for the rest
        while queue:
            current = queue.pop(0)
            for char, node in current.goto.items():
                queue.append(node)
                failure = current.fail
                while failure and char not in failure.goto:
                    failure = failure.fail
                if not failure:
                    node.fail = self.root
                else:
                    node.fail = failure.goto[char]
                    # Add output patterns from failure node
                    node.out.extend(node.fail.out)

        self.built = True

    def search(self, text: Union[str, bytes]) -> Dict[Any, List[int]]:
        """
        Search for patterns in the text and return matching pattern IDs with their positions.

        Args:
            text: A string or bytes to search in

        Returns:
            A dictionary mapping pattern IDs to lists of positions where prefixes were found
        """
        if not self.built:
            self.build_automaton()

        # Ensure text is in the correct type (str or bytes)
        if self.is_bytes and isinstance(text, str):
            text = text.encode(self.encoding)
        elif not self.is_bytes and isinstance(text, bytes):
            text = text.decode(self.encoding, errors="ignore")

        node = self.root
        results = {}

        for i, char in enumerate(text):
            # For bytes, char will be int value of the byte
            char = text[i]

            while node is not self.root and char not in node.goto:
                node = node.fail

            if char in node.goto:
                node = node.goto[char]
                if node.out:
                    for pattern_id in node.out:
                        if pattern_id not in results:
                            results[pattern_id] = []
                        # Store the start position by subtracting the node depth
                        prefix_length = len(self.pattern_prefixes.get(pattern_id, ""))
                        if self.is_bytes and isinstance(
                            prefix := self.pattern_prefixes.get(pattern_id, ""), str
                        ):
                            prefix_length = len(prefix.encode(self.encoding))

                        # Calculate the start position of the match
                        start_pos = max(0, i + 1 - prefix_length)
                        results[pattern_id].append(start_pos)

        return results


def build_regex_literal_matcher(
    patterns_dict: Dict[Any, str],
    is_bytes: bool = False,
    encoding: str = "utf-8",
    is_literal: bool = False,
) -> AhoCorasick:
    """
    Build an Aho-Corasick automaton for matching regex literal strings, typically prefixes.

    Args:
        patterns_dict: Dictionary mapping pattern IDs to regex patterns
        is_bytes: Flag indicating if the automaton should work with bytes
        encoding: Character encoding to use for string/bytes conversion (default: utf-8)
        is_literal: Flag indicating if the patterns should be treated as regex literals (default: False)

    Returns:
        An Aho-Corasick automaton for prefix matching
    """
    ac = AhoCorasick(is_bytes=is_bytes, encoding=encoding)

    for pattern_id, pattern in patterns_dict.items():
        # For fixed literals, we can directly add them to the automaton
        if is_literal:
            ac.add_pattern(pattern, pattern_id, pattern)
            continue
        literals, is_prefix = extract_fixed_literals(pattern)
        if len(literals) > 10:  # Skip excessive number of prefixes
            continue
        if (
            len(literals) == 1 and len(literals[0]) == 1
        ):  # Skip prefix if it just matches a single character
            continue
        for prefix in literals:
            if not prefix:  # Skip empty prefixes
                continue

            ac.add_pattern(prefix, pattern_id, prefix)

    ac.build_automaton()
    return ac



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re
import sys
from typing import List, Tuple

from loguru import logger

# Choose the appropriate module based on Python version
# Note that this is "undocumented" so could break -- may want to find/write an alternative library
if sys.version_info < (3, 11):
    # pylint: disable=deprecated-module
    import sre_parse as re_parser
else:
    import re._parser as re_parser


# pylint: disable=too-many-return-statements
def handle_escaped_literal(regex_pattern: str, i: int, length: int) -> Tuple[str, int]:
    """
    Handles escaped literals in a regex pattern and returns the character to be added to the prefix
    and the number of positions to advance the index by.

    Parameters:
    regex_pattern (str): The regex pattern being parsed.
    i (int): The current index in the regex pattern.
    length (int): The total length of the regex pattern.

    Returns:
    Tuple[str, int]: A tuple containing the character to be added to the prefix and the number of positions
           to advance the index by. Returns (None, 0) if parsing should stop.
    """
    next_char = regex_pattern[i + 1]
    # Try to catch all potential escape sequences with non-literal meaning
    # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_expressions/Cheatsheet
    # https://en.wikipedia.org/wiki/Regular_expression
    if next_char in "dDsSwWbBpPUukAaeZ0123456789":
        return None, 0
    if next_char in "rntvf0":
        escape_sequences = {"r": "\r", "n": "\n", "t": "\t", "v": "\v", "f": "\f", "0": "\0"}
        return escape_sequences[next_char], 2
    if next_char == "c":
        if (
            i + 2 < length
            and regex_pattern[i + 1] == "c"
            and regex_pattern[i + 2].isalpha()
            and regex_pattern[i + 2].isupper()
        ):
            control_char = chr(ord(regex_pattern[i + 2]) - ord("A") + 1)
            return control_char, 3
        return None, 0
    if next_char == "x":
        if i + 3 < length and all(
            c in "0123456789abcdefABCDEF" for c in regex_pattern[i + 2 : i + 4]
        ):
            return chr(int(regex_pattern[i + 2 : i + 4], 16)), 4
        return None, 0
    # Should be able to handle '\uhhhh' case, however there is an additional \u{hhhh} and \u{hhhhh} syntax
    # that may be used if a flag is enabled, which complicates things a bit
    return next_char, 2


def extract_fixed_prefixes(regex_pattern: str) -> List[str]:
    """
    Extracts the fixed string prefixes from a regex pattern, including handling
    escaped characters, expanding character classes, and handling capture groups
    with multiple prefixes separated by '|'.

    Parameters:
    regex_pattern (str): The regex pattern from which to extract the prefixes.

    Returns:
    list: A list of fixed string prefixes, or an empty list if no fixed prefix exists.
    """
    # Compile the regex pattern to ensure it's valid
    try:
        re.compile(regex_pattern)
    except re.error as e:
        raise ValueError(f"Invalid regex pattern: {e}") from None

    prefixes = []
    i = 0
    length = len(regex_pattern)

    # Ignore starting '^'
    if i < length and regex_pattern[i] == "^":
        i += 1

    # Handle character class at the start
    if i < length and regex_pattern[i] == "[":
        i += 1
        char_class = []
        if i < length and regex_pattern[i] == "^":
            # If the character class is negated, we can't handle it -- return an empty prefix list
            return []
        while i < length and regex_pattern[i] != "]":
            if regex_pattern[i] == "\\" and i + 1 < length:
                if regex_pattern[i + 1] == "b":
                    char_class.append("\b")
                    i += 2
                else:
                    next_char, i_delta = handle_escaped_literal(regex_pattern, i, length)
                    # If it's a special character class, we can't handle it -- return empty prefix list
                    if next_char is None:
                        return []
                    char_class.append(next_char)
                    i += i_delta
            elif (
                regex_pattern[i] == "-"
                and char_class
                and i + 1 < length
                and regex_pattern[i + 1] != "]"
            ):
                # A hyphen means we need to handle a range of prefixes from the previous character to the next one
                start_char = char_class.pop()
                end_char = regex_pattern[i + 1]
                char_class.extend(chr(c) for c in range(ord(start_char), ord(end_char) + 1))
                i += 2
            else:
                char_class.append(regex_pattern[i])
                i += 1
        i += 1  # Skip the closing ']'
        # Generate prefixes for each character in the class
        for char in char_class:
            prefixes.append(char)
    # Handle capture group at the start
    elif i < length and regex_pattern[i] == "(":
        i += 1
        if i < length and regex_pattern[i] == "?":
            return []  # Return early if it starts with a '?'
        group_content = []
        while i < length and regex_pattern[i] != ")":
            if regex_pattern[i] == "\\" and i + 1 < length:
                next_char, i_delta = handle_escaped_literal(regex_pattern, i, length)
                if next_char is None:
                    return []  # special character class, we can't handle this form of regex
                group_content.append(next_char)
                i += i_delta
            elif regex_pattern[i] == "(":
                return []  # Return early if it contains a nested capture group
            else:
                group_content.append(regex_pattern[i])
                i += 1
        i += 1  # Skip the closing ')'
        # Split the group content by '|'
        for part in "".join(group_content).split("|"):
            # Only consider alphanumeric prefixes
            if part.isalnum():
                prefixes.append(part)

    # Handle regular fixed prefix extraction
    prefix = []
    while i < length:
        char = regex_pattern[i]
        if char == "\\" and i + 1 < length:
            next_char, i_delta = handle_escaped_literal(regex_pattern, i, length)
            if next_char is None:
                break
            prefix.append(next_char)
            i += i_delta
        elif char.isalnum() or char in "_-,: '":
            prefix.append(char)
            i += 1
        else:
            break
    if prefix:
        prefixes = [p + "".join(prefix) for p in prefixes] if prefixes else ["".join(prefix)]

    return prefixes


# pylint: disable=too-few-public-methods
class RegexNode:
    """A node in the regex parse tree."""

    def __init__(self, op, value=None, children=None):
        self.op = op  # Operation type (LITERAL, BRANCH, etc.)
        self.value = value  # Value for this node (e.g., character for LITERAL)
        self.children = children if children is not None else []

    def __repr__(self):
        if self.value is not None:
            return f"RegexNode({self.op}, {repr(self.value)})"
        return f"RegexNode({self.op}, children={len(self.children)})"


def extract_fixed_literals(pattern, max_possibilities=10, min_length=3):
    """
    Extract fixed literals from a regular expression pattern.

    Args:
        pattern: The regular expression pattern to analyze.
        max_possibilities: Maximum number of different possibilities to generate.
        min_length: Minimum length threshold for a potential fixed string.

    Returns:
        A tuple containing:
        - A list of fixed literals that could be matched by the pattern
        - A boolean indicating whether the patterns are prefixes (True) or not (False)
    """
    try:
        # Parse the regex pattern
        parsed = re_parser.parse(pattern)

        # Convert to our tree structure
        regex_tree = build_regex_tree(parsed, re_parser)

        # First try to extract literal prefixes from the tree
        literals, is_prefix, _is_comprehensive = extract_prefix_from_tree(
            regex_tree, max_possibilities, min_length, re_parser
        )

        if literals:
            return literals, is_prefix

        # Try to find fixed internal literals from the tree
        literals = extract_internal_literals(regex_tree, max_possibilities, min_length, re_parser)
        return literals, False

    except (re.error, AttributeError, TypeError) as e:
        logger.error(f"Error processing regex pattern: {e}")
        return [], False


def build_regex_tree(parsed, parser):
    """Convert the parsed regex into a tree structure."""
    root = RegexNode("ROOT", children=[])

    for op, av in parsed:
        if op == parser.LITERAL:
            # Simple literal character
            root.children.append(RegexNode(op, value=chr(av)))

        elif op == parser.IN:
            # Character class [...]
            chars = extract_chars_from_class(av, parser)
            node = RegexNode(op, value=chars)
            root.children.append(node)

        elif op == parser.SUBPATTERN:
            # Subpattern (...)
            # Structure varies by Python version
            subpattern = av[3] if len(av) > 3 else av[1]
            subtree = build_regex_tree(subpattern, parser)
            node = RegexNode(op, children=subtree.children)
            root.children.append(node)

        elif op == parser.BRANCH:
            # Alternation a|b|c
            branches = av[1]
            branch_nodes = []

            for branch in branches:
                branch_tree = build_regex_tree(branch, parser)
                branch_nodes.append(branch_tree)

            node = RegexNode(op, children=branch_nodes)
            root.children.append(node)

        elif op == parser.MAX_REPEAT:
            # Repetition a*, a+, a{n,m}
            min_count, max_count, subpattern = av
            subtree = build_regex_tree(subpattern, parser)

            node = RegexNode(op, value=(min_count, max_count), children=[subtree])
            root.children.append(node)

        elif op == parser.AT:
            # Anchors (^, $)
            # We can ignore these for fixed prefix extraction
            continue

        else:
            # Any other operation (ANY, etc.)
            node = RegexNode(op, value=av)
            root.children.append(node)

    return root


def extract_chars_from_class(char_class, parser):
    """Extract characters from a character class."""
    chars = set()

    for op, av in char_class:
        if op == parser.LITERAL:
            # Single characters
            chars.add(chr(av))
        elif op == parser.RANGE:
            # Character ranges (e.g., a-z, 0-9)
            start, end = av
            for i in range(start, end + 1):
                chars.add(chr(i))

    return list(chars)


def extract_prefix_from_tree(node, max_possibilities, min_length, parser):
    """Extract literal prefixes from the regex tree."""
    if node.op == "ROOT":
        # Process child nodes sequentially for prefixes
        prefixes = []
        total_possibilities = 1
        is_prefix = True
        is_comprehensive = True

        for child in node.children:
            # Check if we can still continue with a fixed prefix
            child_prefixes, child_is_prefix, is_comprehensive = node_to_prefixes(
                child, max_possibilities // total_possibilities, parser
            )

            if not child_is_prefix:
                break

            if not child_prefixes:
                is_prefix = False
                break

            # If this would create too many possibilities, stop
            if len(child_prefixes) * total_possibilities > max_possibilities:
                is_prefix = False
                break

            # Combine current prefixes with child prefixes
            if not prefixes:
                prefixes = child_prefixes
            else:
                new_prefixes = []
                for prefix in prefixes:
                    for child_prefix in child_prefixes:
                        new_prefixes.append(prefix + child_prefix)
                prefixes = new_prefixes
            total_possibilities *= len(child_prefixes)

            # If the set of child prefixes returned was not comprehensive, we can't keep going
            if not is_comprehensive:
                is_comprehensive = False
                break

        # Filter by minimum length
        prefixes = [p for p in prefixes if len(p) >= min_length]
        return prefixes[:max_possibilities], is_prefix, is_comprehensive

    return [], False, True


def node_to_prefixes(node, max_possibilities, parser):
    """Convert a node to its potential literal prefixes."""
    is_comprehensive = True

    if node.op == parser.LITERAL:
        # Single literal character
        return [node.value], True, is_comprehensive

    if node.op == parser.IN:
        # Character class
        if len(node.value) > max_possibilities:
            return [], False, is_comprehensive
        return node.value, True, is_comprehensive

    if node.op == parser.SUBPATTERN:
        # Process children nodes sequentially, just like we do for ROOT nodes
        prefixes = [""]
        total_possibilities = 1
        is_fixed_prefix = True

        for child in node.children:
            # Get potential prefixes for this child
            child_prefixes, child_is_prefix, is_comprehensive = node_to_prefixes(
                child, max_possibilities // total_possibilities, parser
            )

            # If this child isn't a prefix generator, we're done collecting
            if not child_is_prefix:
                is_fixed_prefix = False
                break

            # If this would create too many possibilities
            if not child_prefixes or (
                len(child_prefixes) * total_possibilities > max_possibilities
            ):
                # If we already have something, return what we've built so far
                if prefixes != [""]:
                    return prefixes, True, is_comprehensive
                # Otherwise, we can't extract a useful prefix
                return [], False, is_comprehensive

            # Combine current prefixes with child prefixes
            new_prefixes = []
            for prefix in prefixes:
                for child_prefix in child_prefixes:
                    new_prefixes.append(prefix + child_prefix)

            prefixes = new_prefixes
            total_possibilities *= len(child_prefixes)

            # If the set of child prefixes returned was not comprehensive, we can't keep going
            if not is_comprehensive:
                is_comprehensive = False
                break

        return prefixes, is_fixed_prefix, is_comprehensive

    if node.op == parser.BRANCH:
        # Alternation - collect prefixes from all branches
        all_prefixes = []
        all_are_prefixes = True

        for branch in node.children:
            branch_prefixes, is_prefix, branch_is_comprehensive = extract_prefix_from_tree(
                branch, max_possibilities, 0, parser
            )

            # If any of the branch prefixes aren't comprehensive, we can't be comprehensive
            if not branch_is_comprehensive:
                is_comprehensive = False

            if not is_prefix:
                all_are_prefixes = False

            all_prefixes.extend(branch_prefixes)

        if len(all_prefixes) > max_possibilities:
            return [], False, is_comprehensive

        return all_prefixes, all_are_prefixes, is_comprehensive

    if node.op == parser.MAX_REPEAT:
        # Repetition
        min_count, max_count = node.value

        # Get prefixes for the repeated content
        sub_tree = node.children[0]
        sub_prefixes, is_prefix, sub_is_comprehensive = extract_prefix_from_tree(
            sub_tree, max_possibilities, 0, parser
        )

        if not is_prefix or not sub_prefixes:
            return [], False, is_comprehensive

        # If max_count is infinite, bound it to the minimum since there's no point in going futher
        # later parts of the pattern can't safely add anything to the prefixes returned anyway
        if str(max_count) == "MAXREPEAT":
            max_count = min_count
            is_comprehensive = False

        # Create the list of prefixes
        prefixes = []

        # If min_count is 0, then the empty prefix will be included (the empty prefix on its own doesn't bring us closer to max_possibilities)
        # Otherwise we'll build up the guaranteed prefix as close to min_count repetitions as possible
        guaranteed_prefixes = [""]
        # If the subtree is comprehensive, we can build up the guaranteed prefixes
        if sub_is_comprehensive:
            for _ in range(min_count):
                new_prefixes = []
                for prefix in guaranteed_prefixes:
                    for sub_prefix in sub_prefixes:
                        new_prefixes.append(prefix + sub_prefix)
                # If we exceed the max possibilities, flag as not comprehensive and break
                if len(new_prefixes) > max_possibilities:
                    is_comprehensive = False
                    break
                # Adding this additional repetition stays within the max possibilities
                guaranteed_prefixes = new_prefixes
        else:
            # If the subtree is not comprehensive, we can only add the prefixes as they are
            guaranteed_prefixes = sub_prefixes
            is_comprehensive = False
        prefixes.extend(guaranteed_prefixes)

        # Next, we can try to insert additional prefix variations up to the max if the prefixes will still be comprehensive
        # If we can reach the max, the additional prefixes can be included
        # Otherwise, there is no point in adding them and we should set is_comprehensive to False
        if is_comprehensive:
            not_guaranteed_prefixes = []
            prev_new_prefixes = prefixes
            for _ in range(max_count - min_count):
                new_prefixes = []
                for prefix in prev_new_prefixes:
                    for sub_prefix in sub_prefixes:
                        new_prefixes.append(prefix + sub_prefix)
                # If we exceed the max possibilities, flag as not comprehensive and break
                if (
                    len(new_prefixes) + len(not_guaranteed_prefixes) + len(prefixes)
                    > max_possibilities
                ):
                    is_comprehensive = False
                    break
                not_guaranteed_prefixes.extend(new_prefixes)
                prev_new_prefixes = new_prefixes
            # Adding these additional repetitions will stay within the max possibilities if its still comprehensive
            if is_comprehensive:
                prefixes.extend(not_guaranteed_prefixes)

        return prefixes, True, is_comprehensive

    # Any other type of node breaks a fixed prefix
    return [], False, is_comprehensive


def extract_internal_literals(node, max_possibilities, min_length, parser):
    """
    Extract fixed internal literals from the regex tree.
    This function identifies internal sequences that must always be present.
    """
    if node.op == "ROOT":
        # First, try to find fixed literals at the top level
        candidates = []

        # Look for consecutive LITERAL nodes
        current_literal = ""
        for child in node.children:
            if child.op == parser.LITERAL:
                current_literal += child.value
            else:
                if len(current_literal) >= min_length:
                    candidates.append(current_literal)
                current_literal = ""

                # Recursively check this non-literal node
                child_literals = extract_internal_literals(
                    child, max_possibilities, min_length, parser
                )
                candidates.extend(child_literals)

        # Check the last literal sequence
        if len(current_literal) >= min_length:
            candidates.append(current_literal)

        # Remove duplicates while preserving order
        seen = set()
        unique_literals = []
        for literal in candidates:
            if literal not in seen:
                seen.add(literal)
                unique_literals.append(literal)

        return unique_literals[:max_possibilities]

    if node.op == parser.SUBPATTERN:
        # Check inside subpatterns
        return extract_internal_literals(
            RegexNode("ROOT", children=node.children), max_possibilities, min_length, parser
        )

    if node.op == parser.BRANCH:
        # For branches, find literals common to all alternatives
        common_literals = []

        for branch in node.children:
            branch_literals = extract_internal_literals(
                branch, max_possibilities, min_length, parser
            )

            # For the first branch, initialize common_literals
            if not common_literals and branch_literals:
                common_literals = branch_literals
            else:
                # Keep only literals that appear in all branches
                common_literals = [
                    lit for lit in common_literals if any(lit in b_lit for b_lit in branch_literals)
                ]

        return common_literals

    if node.op == parser.MAX_REPEAT:
        min_count = node.value[0]

        # Only consider required repetitions
        if min_count > 0:
            sub_tree = node.children[0]
            sub_literals = extract_internal_literals(
                sub_tree, max_possibilities, min_length, parser
            )

            # If always repeated at least once, these literals must be present
            return sub_literals

    return []



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import gzip
import subprocess
import tempfile
from typing import List, Optional

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def check_if_grype_installed() -> bool:
    try:
        result = subprocess.run(["grype", "--help"], capture_output=True, check=False).returncode
        if result == 0:
            logger.debug("Grype is installed and ready to use.")
        else:
            logger.warning("Grype is not installed or not functioning correctly.")
        return result == 0
    except FileNotFoundError:
        logger.error("Grype is not installed. Please install it to use this plugin.")
        return False


disable_plugin = not check_if_grype_installed()


def run_grype(filename: str) -> object:
    logger.debug(f"Starting grype scan for file: {filename}")
    result = subprocess.run(["grype", filename], capture_output=True, check=False)
    if result.returncode != 0:
        logger.warning(f"Running grype on {filename} failed with return code {result.returncode}")
        logger.debug(f"Grype stderr: {result.stderr.decode().strip()}")
        return None
    logger.debug(f"Grype scan completed successfully for file: {filename}")
    output = result.stdout.decode()
    logger.debug(
        f"Grype raw output: {output[:200]}..."
    )  # Log the first 200 characters of the output
    to_ret = []
    # skip the header on the first line
    for line in output.split("\n")[1:]:
        columns = [s.strip() for s in line.split("  ") if s.strip()]
        # Skip empty lines
        if len(columns) == 0:
            continue
        # Assume that the "Fixed In" field is missing if there's only 5 entries
        name = columns[0]
        installed = columns[1]
        if len(columns) == 5:
            fixed_in = ""
            type_ = columns[2]
            vuln = columns[3]
            severity = columns[4]
        else:
            fixed_in = columns[2]
            type_ = columns[3]
            vuln = columns[4]
            severity = columns[5]
        to_ret.append(
            {
                "name": name,
                "installed": installed,
                "fixed_in": fixed_in,
                "type": type_,
                "vulnerability": vuln,
                "severity": severity,
            }
        )
    return {"grype_output": to_ret}


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM, software: Software, filename: str, filetype: List[str], children: list
) -> Optional[List[Software]]:
    if disable_plugin:
        return None

    if not ("DOCKER_TAR" in filetype or "DOCKER_GZIP" in filetype):
        return None

    try:
        if "DOCKER_GZIP" in filetype:
            logger.debug(f"Decompressing gzip file: {filename}")
            with open(filename, "rb") as gzip_in:
                gzip_data = gzip_in.read()
            with tempfile.NamedTemporaryFile() as gzip_out:
                gzip_out.write(gzip.decompress(gzip_data))
                logger.debug(f"Decompressed file written to temporary file: {gzip_out.name}")
                return run_grype(gzip_out.name)

        # For DOCKER_TAR or other supported types
        return run_grype(filename)

    except FileNotFoundError:
        logger.error(f"File not found: {filename}")
    except PermissionError:
        logger.error(f"Permission denied when accessing file: {filename}")
    except gzip.BadGzipFile:
        logger.error(f"Invalid gzip file: {filename}")
    except OSError as e:
        logger.error(f"OS error while processing file {filename}: {e}")

    return None


@surfactant.plugin.hookimpl
def update_db():
    logger.info("Starting grype database update...")
    # Example update logic
    try:
        subprocess.check_call(["grype", "db", "update"])
        logger.info("Grype database updated successfully.")
        return "Database updated successfully."
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to update Grype database: {e}")
        return f"Failed to update database: {e}"


@surfactant.plugin.hookimpl
def short_name():
    return "grype"



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import re
from typing import Any, Dict, List, Optional, Set, Union

from loguru import logger

import surfactant.plugin
from surfactant.database_manager.database_utils import BaseDatabaseManager, DatabaseConfig
from surfactant.sbomtypes import SBOM, Software
from surfactant.utils.ahocorasick import build_regex_literal_matcher

# Global configuration
DATABASE_URL_EMBA = "https://raw.githubusercontent.com/e-m-b-a/emba/11d6c281189c3a14fc56f243859b0bccccce8b9a/config/bin_version_strings.cfg"
NATIVE_DB_DIR = "native_library_patterns"  # The directory name to store the database toml file and database json files for this module


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "native_lib_file"


class EmbaNativeLibDatabaseManager(BaseDatabaseManager):
    """Manages the EMBA Native Library database."""

    def __init__(self):
        name = (
            short_name()
        )  # Set to '__name__' (without quotation marks), if short_name is not implemented

        config = DatabaseConfig(
            database_dir=NATIVE_DB_DIR,  # The directory name to store the database toml file and database json files for this module.
            database_key="emba",  # The key for this classes database in the version_info toml file.
            database_file="emba_db.json",  # The json file name for the database.
            source=DATABASE_URL_EMBA,  # The source of the database (put "file" or the source url)
            plugin_name=name,
        )

        super().__init__(config)
        self.ac_filename = None
        self.ac_filecontent = None

    def parse_raw_data(self, raw_data: str) -> Dict[str, Dict[str, List[str]]]:
        database: Dict[str, Dict[str, List[str]]] = {}
        lines = raw_data.splitlines()
        filtered_lines: List[str] = []

        for line in lines:
            if not line.startswith("#"):
                filtered_lines.append(line)

        for line in filtered_lines:
            line = line.strip()

            fields = line.split(";")

            lib_name = fields[0]

            name_patterns: List[str] = []

            if fields[3].startswith('"') and fields[3].endswith('""'):
                filecontent = fields[3][1:-1]
            elif fields[3].endswith('""'):
                filecontent = fields[3][:-1]
            else:
                filecontent = fields[3].strip('"')

            if fields[1] == "" or fields[1] == "strict":
                if fields[1] == "strict":
                    if lib_name not in database:
                        database[lib_name] = {
                            "filename": [lib_name],
                            "filecontent": [],
                        }
                    else:
                        if lib_name not in database[lib_name]["filename"]:
                            database[lib_name]["filename"].append(lib_name)
                else:
                    try:
                        re.compile(filecontent.encode("utf-8"))  # Validate regex
                        if lib_name not in database:
                            database[lib_name] = {
                                "filename": name_patterns,
                                "filecontent": [filecontent],
                            }
                        else:
                            database[lib_name]["filecontent"].append(filecontent)

                    except re.error as rex:
                        logger.error("Error parsing file content regexp %s: %s", filecontent, rex)
        return database

    def load_db(self) -> Optional[Dict[str, Any]]:
        """Load the database and build the Aho-Corasick automaton for pattern matching."""
        super().load_db()
        if self._database:
            # Build the Aho-Corasick automaton for filename patterns
            filename_patterns = {}
            filecontent_patterns = {}

            for lib_name, lib_data in self._database.items():
                if "filename" in lib_data:
                    for pattern in lib_data["filename"]:
                        pattern_id = (lib_name, "filename", pattern)
                        filename_patterns[pattern_id] = pattern

                if "filecontent" in lib_data:
                    for pattern in lib_data["filecontent"]:
                        pattern_id = (lib_name, "filecontent", pattern)
                        filecontent_patterns[pattern_id] = pattern.encode("utf-8").decode(
                            "unicode_escape"
                        )

            if filename_patterns:
                self.ac_filename = build_regex_literal_matcher(filename_patterns, is_literal=True)

            if filecontent_patterns:
                self.ac_filecontent = build_regex_literal_matcher(
                    filecontent_patterns, is_bytes=True
                )
        return self._database


native_lib_manager = EmbaNativeLibDatabaseManager()


def supports_file(filetype: List[str]) -> bool:
    supported_types = ("PE", "ELF", "MACHOFAT", "MACHOFAT64", "MACHO32", "MACHO64", "UIMAGE")
    for ft in filetype:
        if ft in supported_types:
            return True
    return False


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM, software: Software, filename: str, filetype: List[str]
) -> Optional[Dict[str, Any]]:
    if not supports_file(filetype):
        return None
    return extract_native_lib_info(filename)


def extract_native_lib_info(filename: str) -> Optional[Dict[str, Any]]:
    native_lib_info: Dict[str, Any] = {"nativeLibraries": []}
    native_lib_database = native_lib_manager.get_database()

    if native_lib_database is None:
        return None

    found_libraries: set = set()
    library_names: List[str] = []
    contains_library_names: List[str] = []

    base_filename = os.path.basename(filename)
    filenames_list = match_by_attribute("filename", base_filename, native_lib_database)
    if len(filenames_list) > 0:
        for match in filenames_list:
            library_name = match["isLibrary"]
            if library_name not in found_libraries:
                library_names.append(library_name)
                found_libraries.add(library_name)

    try:
        with open(filename, "rb") as native_file:
            filecontent = native_file.read()
        filecontent_list = match_by_attribute("filecontent", filecontent, native_lib_database)

        for match in filecontent_list:
            library_name = match["containsLibrary"]
            if library_name not in found_libraries:
                contains_library_names.append(library_name)
                found_libraries.add(library_name)

    except FileNotFoundError:
        logger.warning("File not found: %s", filename)

    if library_names:
        native_lib_info["nativeLibraries"].append({"isLibrary": library_names})

    if contains_library_names:
        native_lib_info["nativeLibraries"].append({"containsLibrary": contains_library_names})

    return native_lib_info


def match_by_attribute(
    attribute: str, content: Union[str, bytes], patterns_database: Dict[str, Any]
) -> List[Dict[str, Any]]:
    libs: List[Dict[str, str]] = []

    # Get the appropriate automaton
    if attribute == "filename":
        ac = native_lib_manager.ac_filename
    elif attribute == "filecontent":
        ac = native_lib_manager.ac_filecontent
    else:
        return libs

    # If no automaton available, fall back to old method
    if not ac:
        for lib_name, lib_info in patterns_database.items():
            if attribute in lib_info:
                for pattern in lib_info[attribute]:
                    if attribute == "filename":
                        if pattern.lower() == content.lower():
                            libs.append({"isLibrary": lib_name})

                    elif attribute == "filecontent":
                        matches = re.search(pattern.encode("utf-8"), content)
                        if matches:
                            libs.append({"containsLibrary": lib_name})
        return libs

    # Use Aho-Corasick for prefix matching
    matched_libraries: Set[str] = set()

    # For filename we can just do direct string matching
    if attribute == "filename":
        if isinstance(content, str):
            potential_matches = ac.search(content)
            for pattern_id, _positions in potential_matches.items():
                lib_name, attr, pattern = pattern_id
                if (
                    attr == attribute
                    and pattern.lower() == content.lower()
                    and lib_name not in matched_libraries
                ):
                    libs.append({"isLibrary": lib_name})
                    matched_libraries.add(lib_name)

    # For file content, we need to search in binary data
    elif attribute == "filecontent":
        if isinstance(content, bytes):
            # Window size for context around matches
            window_size = 4096  # Adjustable based on expected pattern size
            content_length = len(content)
            potential_matches = ac.search(content)
            for pattern_id, positions in potential_matches.items():
                lib_name, attr, pattern = pattern_id
                if attr == attribute and lib_name not in matched_libraries:
                    try:
                        encoded_pattern = pattern.encode("utf-8")

                        # Check each position where the prefix was found
                        for pos in positions:
                            # Calculate start and end for a slice of content around the match position
                            slice_start = max(0, pos - 50)  # Allow some context before match
                            slice_end = min(content_length, pos + window_size)
                            content_slice = content[slice_start:slice_end]

                            # Search only in the slice
                            matches = re.search(encoded_pattern, content_slice)
                            if matches:
                                libs.append({"containsLibrary": lib_name})
                                matched_libraries.add(lib_name)
                                break  # Found one match for this library, no need to check more positions
                    except re.error as e:
                        logger.error(f"Error with regex pattern {pattern}: {e}")

    return libs


@surfactant.plugin.hookimpl
def update_db() -> str:
    return native_lib_manager.download_and_update_database()


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None) -> None:
    """
    Initialization hook to load the native library database.

    Args:
        command_name (Optional[str], optional): The name of the command invoking the initialization.
            If set to "update-db", the database will not be loaded.

    Returns:
        None
    """
    if command_name != "update-db":
        logger.info("Initializing native_lib_file...")
        native_lib_manager.load_db()
        logger.info("Initializing native_lib_file complete.")



#!/usr/bin/env python3
"""
Oneshot utility to generate SBOM from a single input folder and return as string.
"""

import argparse
import datetime
import difflib
import io
import json
import os
import random
import sys
import time
import traceback
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import Optional
from unittest.mock import patch

import click
from loguru import logger

from surfactant.cmd.generate import sbom


def deterministic_uuid4():
    return str(uuid.UUID(bytes=random.randbytes(16), version=4))


def deterministic_time():
    return 1609459200  # Friday, January 1, 2021 12:00:00 AM UTC


_json_dumps = json.dumps


def deterministic_json_dumps(*args, **kwargs):
    kwargs["sort_keys"] = True
    return _json_dumps(*args, **kwargs)


_json_dump = json.dump


def deterministic_json_dump(*args, **kwargs):
    kwargs["sort_keys"] = True
    return _json_dump(*args, **kwargs)


class FixedDateTime(datetime.datetime):
    @classmethod
    def now(cls, tz=None):
        return cls.fromtimestamp(time.time(), tz)

    @classmethod
    def utcnow(cls):
        return cls.fromtimestamp(time.time(), datetime.timezone.utc)


@contextmanager
def deterministic_context(enabled: bool = True):
    """Context manager to optionally patch time-related and random functions for deterministic output.

    Args:
        enabled (bool): If True, applies patches for deterministic behavior. If False, no patches are applied.
    """
    if enabled:
        with (
            patch("uuid.uuid4", side_effect=deterministic_uuid4),
            patch("datetime.datetime", FixedDateTime),
            patch("time.time", side_effect=deterministic_time),
            patch("json.dumps", side_effect=deterministic_json_dumps),
            patch("json.dump", side_effect=deterministic_json_dump),
        ):
            # Set the random seed for deterministic random number generation
            random.seed(0xDEADBEEF)
            yield
    else:
        # No patches applied, just yield normally
        yield


def generate_sbom_string(
    input_folder: str,
    install_prefix: Optional[str] = None,
    deterministic: bool = False,
) -> str:
    """
    Generate an SBOM from a single input folder and return it as a string.

    Args:
        input_folder (str): Path to the folder to analyze
        install_prefix (Optional[str]): Install prefix for the software. If None, uses the folder path.
        deterministic (bool): Use deterministic UUIDs and timestamps for reproducible output

    Returns:
        str: The generated SBOM as a string

    Raises:
        FileNotFoundError: If the input folder doesn't exist
        ValueError: If the input folder is not a directory
    """
    # Validate input folder
    folder_path = Path(input_folder)
    if not folder_path.exists():
        raise FileNotFoundError(f"Input folder does not exist: {input_folder}")
    if not folder_path.is_dir():
        raise ValueError(f"Input path is not a directory: {input_folder}")

    # Set install prefix if not provided
    if install_prefix is None:
        install_prefix = folder_path.as_posix() + "/"

    # Create specimen context for the single folder
    specimen_context = [{"extractPaths": [folder_path.as_posix()], "installPrefix": install_prefix}]

    # Create an in-memory file-like object to capture output
    output_buffer = io.StringIO()

    # Create a click context
    with click.Context(sbom) as ctx:
        with deterministic_context(enabled=deterministic):
            try:
                # Use Click's invoke to call the command with the context
                ctx.invoke(
                    sbom,
                    specimen_context=specimen_context,
                    sbom_outfile=output_buffer,
                )
            except Exception as e:
                raise RuntimeError(f"Failed to invoke SBOM generation: {e}") from e

    # Get the output as a string
    return output_buffer.getvalue()


data_path = Path(__file__).parent.parent / "tests" / "data"
data_folders = []
if not data_path.exists():
    logger.error(f"Test data directory does not exist: {data_path}")
else:
    data_folders = [item for item in data_path.iterdir() if item.is_dir()]


def test_all_data_folders():
    """
    Test the generate_sbom_string function with all folders in tests/data.
    """
    if not data_folders:
        logger.warning("No test data folders found")
        return

    logger.info(f"Testing SBOM generation for {len(data_folders)} folders")

    for folder in data_folders:
        logger.info(f"Testing folder: {folder.name}")

        try:
            # Test regular mode
            sbom_string = generate_sbom_string(
                input_folder=str(folder),
                deterministic=False,
            )

            # Test deterministic mode
            sbom_string_det1 = generate_sbom_string(
                input_folder=str(folder),
                deterministic=True,
            )

            # Test deterministic mode
            sbom_string_det2 = generate_sbom_string(
                input_folder=str(folder),
                deterministic=True,
            )

            # Print first few lines of the SBOM to verify it was generated
            if sbom_string and sbom_string_det1 and sbom_string_det2:
                logger.success("SBOM generated successfully")

                # Verify deterministic output is different from regular output
                if sbom_string in (sbom_string_det1, sbom_string_det2):
                    logger.warning("Deterministic and regular output are identical (unexpected)")
                else:
                    logger.success("Deterministic mode produces different output as expected")

                # Check if deterministic outputs are identical
                if sbom_string_det1 == sbom_string_det2:
                    logger.success("Deterministic outputs are identical as expected")
                else:
                    logger.warning("Deterministic outputs are different (unexpected)")
                    # Show differences between the two deterministic runs
                    for line in show_diff(sbom_string_det1, sbom_string_det2).splitlines():
                        logger.info(line)
            else:
                logger.warning("SBOM generated but appears to be empty")

        except (FileNotFoundError, ValueError, RuntimeError) as e:
            logger.error(f"Error generating SBOM for {folder.name}: {e}")


def test_gha(
    old_folders: dict[str, dict[str, Optional[str]]],
    repo: Optional[str],
    current_run: Optional[tuple[str, str]],
    last_run: Optional[tuple[str, str]],
    requested_last_sha: Optional[str] = None,
) -> tuple[str, dict[str, dict[str, Optional[str]]]]:
    """
    Test function for CI/CD mode.

    Args:
        old_folders (dict[str, str]): Dictionary of old folder names and their SBOM strings.
        repo_prefix (Optional[str]): Optional prefix for the repository URL to link to the folders.

    Returns:
        str: A github step summary message.
    """
    folders = [folder.name for folder in data_folders]
    created_folders = [folder for folder in folders if folder not in old_folders]
    removed_folders = [folder for folder in old_folders if folder not in folders]

    summary = ""

    if created_folders:
        logger.info(f"New folders detected: {', '.join(created_folders)}")
        summary += f"## 🆕 New Folders ({len(created_folders)})\n"
        for folder in created_folders:
            summary += f"- {folder}\n"
        summary += "\n"
    if removed_folders:
        logger.info(f"Removed folders: {', '.join(removed_folders)}")
        summary += f"## 🗑️ Removed Folders ({len(removed_folders)})\n"
        for folder in removed_folders:
            summary += f"- {folder}\n"
        summary += "\n"

    logger.info(f"Testing SBOM generation for {len(data_folders)} folders")

    new_folders = {}
    if not data_folders:
        summary += "## ❓ No Test Folders Found\n"
    else:
        results = {}
        for folder in data_folders:
            old_data = old_folders.get(folder.name, {"sbom": "", "stacktrace": None})
            new_data = {"sbom": old_data["sbom"], "stacktrace": None}
            try:
                new_data["sbom"] = generate_sbom_string(
                    input_folder=str(folder),
                    deterministic=True,
                )
            except (FileNotFoundError, ValueError, RuntimeError) as e:
                logger.error(f"Error generating SBOM for {folder.name}: {e}")
                new_data["stacktrace"] = traceback.format_exc()

            new_folders[folder.name] = new_data
            if new_data["sbom"] != old_data["sbom"] and new_data["sbom"]:
                logger.info(f"Changes detected in folder: {folder.name}")
                for line in show_diff(old_data["sbom"], new_data["sbom"]).splitlines():
                    logger.info(line)
                results[folder.name] = {"diff": show_diff(old_data["sbom"], new_data["sbom"], 100)}
            elif new_data["stacktrace"]:
                results[folder.name] = {"stacktrace": new_data["stacktrace"]}

        if not results:
            summary += "## ✅ No SBOM Changes Detected\n"
        else:
            summary += f"## 🧪 SBOM Results ({len(results)}/{len(data_folders)})\n\n"
            for folder_name, result in results.items():
                summary += "<details>\n"
                summary += "<summary><h3>"
                if "stacktrace" in result:
                    summary += "❗️ "
                summary += f"{folder_name}"
                if repo and current_run:
                    href = (
                        f"https://github.com/{repo}/tree/{current_run[0]}/tests/data/{folder_name}"
                    )
                    summary += f' (<a href="{href}">Link</a>)'
                summary += "</h3></summary>\n\n"
                if "stacktrace" in result:
                    summary += f"```\n{result['stacktrace']}\n```\n"
                elif "diff" in result:
                    summary += f"```diff\n{result['diff']}\n```\n"
                summary += "</details>\n"

        if repo and current_run:
            run_href = f"https://github.com/{repo}/actions/runs/{current_run[1]}"
            # commit_href = f"https://github.com/{repo}/commit/{current_run[0]}"
            summary += f"\n<sub>For commit {current_run[0][:7]} (Run <a href='{run_href}'>{current_run[1]}</a>)</sub>"
        if repo and last_run:
            run_href = f"https://github.com/{repo}/actions/runs/{last_run[1]}"
            # commit_href = f"https://github.com/{repo}/commit/{last_run[0]}"
            summary += f"\n<sub>Compared against commit {last_run[0][:7]} (Run <a href='{run_href}'>{last_run[1]}</a>)</sub>"
            if requested_last_sha and requested_last_sha != last_run[0]:
                # commit_href = f"https://github.com/{repo}/commit/{requested_last_sha}"
                summary += f"\n<sub>⚠️ Requested SHA {requested_last_sha[:7]} did not have a matching run. Used {last_run[0][:7]} instead.</sub>"
    return summary.rstrip(), new_folders


def show_diff(text1: str, text2: str, max_lines: int = 20) -> str:
    """
    Show differences between two texts.

    Args:
        text1 (str): First text to compare
        text2 (str): Second text to compare
        max_lines (int): Maximum number of diff lines to show

    Returns:
        str: Formatted string showing the differences
    """
    lines1 = text1.splitlines(keepends=True)
    lines2 = text2.splitlines(keepends=True)

    diff = list(difflib.unified_diff(lines1, lines2, lineterm=""))

    # Show first max_lines lines of diff
    lines = [line.rstrip() for line in diff[:max_lines]]

    if len(diff) > max_lines:
        lines.append(f"... and {len(diff) - max_lines} more lines")

    return "\n".join(lines)


def main():
    """
    Example usage of the generate_sbom_string function.
    """
    parser = argparse.ArgumentParser(description="Generate SBOM from a single folder")
    parser.add_argument("folder", nargs="?", help="Path to the folder to analyze")
    parser.add_argument("--install-prefix", help="Install prefix for the software")
    parser.add_argument(
        "--deterministic",
        action="store_true",
        help="Use deterministic UUIDs and timestamps for reproducible output",
    )
    parser.add_argument(
        "--test-all", action="store_true", help="Test with all folders in tests/data"
    )
    parser.add_argument("--gha", action="store_true", help="CI/CD mode")

    args = parser.parse_args()

    if args.gha:
        logger.info("Running in CI/CD mode o/")
        input_file = os.environ.get("DIFF_INPUT", None)
        output_file = os.environ.get("DIFF_OUTPUT", None)
        summary_file = os.environ.get("SUMMARY_OUTPUT", None)
        repo = os.environ.get("REPO", None)
        current_sha = os.environ.get("CURRENT_RUN_SHA", None)
        current_id = os.environ.get("CURRENT_RUN_ID", None)
        last_sha = os.environ.get("LAST_RUN_SHA", None)
        last_id = os.environ.get("LAST_RUN_ID", None)
        requested_last_sha = os.environ.get("REQUESTED_LAST_SHA", "")
        gh_summary = os.environ.get("GITHUB_STEP_SUMMARY", None)
        old_folders = {}
        if input_file:
            with open(input_file, "r") as f:
                old_folders = json.load(f)
        summary, new_folders = test_gha(
            old_folders,
            repo,
            (current_sha, current_id) if current_sha and current_id else None,
            (last_sha, last_id) if last_sha and last_id else None,
            requested_last_sha,
        )

        if summary_file:
            with open(summary_file, "w") as f:
                print(summary, file=f)
        else:
            print(summary)

        if gh_summary:
            with open(gh_summary, "a") as f:
                print(summary, file=f)

        if output_file:
            with open(output_file, "w") as f:
                json.dump(new_folders, f, indent=4)
        return

    if args.test_all:
        test_all_data_folders()
        return

    if not args.folder:
        logger.error("folder argument is required unless --test-all is specified")
        parser.print_help()
        sys.exit(1)

    try:
        sbom_string = generate_sbom_string(
            input_folder=args.folder,
            install_prefix=args.install_prefix,
            deterministic=args.deterministic,
        )
        print(sbom_string)
    except (FileNotFoundError, ValueError, RuntimeError) as e:
        logger.error(f"Error generating SBOM: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import os
import pathlib
import tarfile
from enum import Enum, auto
from typing import Optional

import surfactant.plugin
from surfactant.infoextractors.coff_file import COFF_MAGIC_TARGET_NAME


class ExeType(Enum):
    ELF = auto()
    PE = auto()
    OLE = auto()
    JAVA_MACHOFAT = auto()
    MACHO32 = auto()
    MACHO64 = auto()


def is_docker_archive(filepath: str) -> bool:
    try:
        # pylint: disable=too-many-return-statements
        with tarfile.open(filepath) as tar:
            try:
                manifest_info = tar.getmember("manifest.json")
                if not manifest_info.isfile():
                    return False
                with tar.extractfile(manifest_info) as manifest_file:
                    manifest = json.load(manifest_file)
                    # There's one entry in the list for each image
                    if not isinstance(manifest, list):
                        return False
                    for data in manifest:
                        # Just check if this data member exists
                        _ = tar.getmember(data["Config"])
                        # Now check that each of the layers exist
                        for layer in data["Layers"]:
                            _ = tar.getmember(layer)
                    # Everything seems to exist and be in order; this is most likely a Docker archive
                    return True
            except KeyError:
                return False
    except tarfile.ReadError:
        return False


@surfactant.plugin.hookimpl(tryfirst=True)
def identify_file_type(filepath: str) -> Optional[str]:
    filetype_matches = []
    try:
        with open(filepath, "rb") as f:
            magic_bytes = f.read(265)
            if magic_bytes[:4] == b"\x7fELF":
                filetype_matches.append("ELF")
            if magic_bytes[:2] == b"MZ":
                # Several file types start with the same `MZ` signature, so we need to handle them.
                # Regardless, the initial header (may) contain a pointer to an additional COFF
                # header, so we look for that as the first step.
                coff_addr = (
                    int.from_bytes(magic_bytes[0x3C:0x40], byteorder="little", signed=False)
                    & 0xFFFF
                )
                # Check to see if the coff_addr is still within the initial read; if not, we read up
                # to where it is.
                if coff_addr > len(magic_bytes):
                    magic_bytes += f.read(coff_addr + 4 - len(magic_bytes))
                # If coff_addr is still longer than what has been read so far, it points off the end
                # of the file, so the file is either malformed or something else is up.
                if coff_addr + 4 > len(magic_bytes):
                    filetype_matches.append("Malformed PE")
                elif magic_bytes[coff_addr : coff_addr + 4] != b"PE\x00\x00":
                    filetype_matches.append("DOS")
                # Check for the linux kernel header at 0x202 (may require a second read)
                else:
                    if len(magic_bytes) < 0x206:
                        magic_bytes += f.read(265)
                    if magic_bytes[0x202:0x206] == b"HdrS":
                        filetype_matches.append("Linux Kernel Image")
                    # Otherwise, call it a PE and be done with it.
                    else:
                        filetype_matches.append("PE")

            if magic_bytes[:8] == b"\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1":
                # MSI (install), MSP (patch), MST (transform), and MSM (merge) files are all types of OLE files
                # the root storage object CLSID is used to identify what it is (+ file extension)
                filetype_matches.append("OLE")
            # Microsoft CAB files
            if magic_bytes[:4] == b"MSCF":
                filetype_matches.append("MSCAB")
            # InstallShield CAB files
            if magic_bytes[:4] == b"ISc(":
                filetype_matches.append("ISCAB")
            # For gzipped data, also filter by extension to avoid huge number of entries with limited info
            if magic_bytes[:2] == b"\x1f\x8b":
                if is_docker_archive(filepath):
                    filetype_matches.append("DOCKER_GZIP")
                else:
                    filetype_matches.append("GZIP")
            # Check for compressed TAR files (tar.bz2, tar.xz)
            if magic_bytes[:3] == b"BZh":
                filetype_matches.append("BZIP2")
            if magic_bytes[:6] == b"\xfd\x37\x7a\x58\x5a\x00":
                filetype_matches.append("XZ")
            if magic_bytes[257:265] == b"ustar\x0000" or magic_bytes[257:265] == b"ustar  \x00":
                if is_docker_archive(filepath):
                    filetype_matches.append("DOCKER_TAR")
                else:
                    filetype_matches.append("TAR")
            if magic_bytes[:6] == b"\x52\x61\x72\x21\x1a\x07":
                filetype_matches.append("RAR")
            if magic_bytes[:4] in [b"PK\x03\x04", b"PK\x05\x06", b"PK\x07\x08"]:
                suffix = pathlib.Path(filepath).suffix.lower()
                if suffix in [".zip", ".zipx"]:
                    filetype_matches.append("ZIP")
                # Java archive files of various types
                if suffix == ".jar":
                    filetype_matches.append("JAR")
                if suffix == ".war":
                    filetype_matches.append("WAR")
                if suffix == ".ear":
                    filetype_matches.append("EAR")
                # Android packages
                if suffix == ".apk":
                    filetype_matches.append("APK")
                # iOS/iPad applications
                if suffix == ".ipa":
                    filetype_matches.append("IPA")
                # Windows app package
                if suffix == ".msix":
                    filetype_matches.append("MSIX")
            # Magic for Java and Mach-O FAT Binary are the same
            if magic_bytes[:4] == b"\xca\xfe\xba\xbe":
                # Distinguish them the same way file magic and Apple do
                # https://opensource.apple.com/source/file/file-80.40.2/file/magic/Magdir/cafebabe.auto.html
                # https://github.com/file/file/blob/master/magic/Magdir/cafebabe
                if int.from_bytes(magic_bytes[4:8], byteorder="big", signed=False) <= 30:
                    filetype_matches.append("MACHOFAT")
                filetype_matches.append("JAVACLASS")
            if magic_bytes[:4] == b"\xbe\xba\xfe\xca":
                filetype_matches.append("MACHOFAT")
            if magic_bytes[:4] in [b"\xca\xfe\xba\xbf", b"\xbf\xba\xfe\xca"]:
                filetype_matches.append("MACHOFAT64")
            # Apple fat EFI binaries
            if magic_bytes[:4] == b"\x0e\xf1\xfa\b9":
                filetype_matches.append("EFIFAT")
            # NOTE: the MACH032 and MACHO64 (normal byte order) magic may be located
            # at offset 0x1000 in some files
            if magic_bytes[:4] in [b"\xfe\xed\xfa\xce", b"\xce\xfa\xed\xfe"]:
                filetype_matches.append("MACHO32")
            if magic_bytes[:4] in [b"\xfe\xed\xfa\xcf", b"\xcf\xfa\xed\xfe"]:
                filetype_matches.append("MACHO64")
            # https://releases.llvm.org/2.7/docs/BitCodeFormat.html#magic
            if magic_bytes[:4] == b"\xde\xc0\x17\x0b":
                filetype_matches.append("LLVM_BITCODE")
            if magic_bytes[:4] == b"BC\xc0\xde":
                filetype_matches.append("LLVM_IR")
            # Need to check both small and big endian for a.out
            a_out_magic = [0x111, 0x108, 0x107, 0x0CC, 0x10B]
            if (
                int.from_bytes(magic_bytes[:4], byteorder="big", signed=False) & 0xFFFF
                in a_out_magic
            ):
                filetype_matches.append("A.OUT big")
            if (
                int.from_bytes(magic_bytes[:4], byteorder="little", signed=False) & 0xFFFF
                in a_out_magic
            ):
                filetype_matches.append("A.OUT little")
            if (
                int.from_bytes(magic_bytes[:2], byteorder="little", signed=False)
                in COFF_MAGIC_TARGET_NAME
            ):
                filetype_matches.append("COFF")
            # XCOFF:
            # https://www.ibm.com/docs/en/aix/7.3?topic=formats-xcoff-object-file-format
            if magic_bytes[:2] == "\x1d\x00":
                filetype_matches.append("XCOFF32")
            if magic_bytes[:2] == "\xf7\x01":
                filetype_matches.append("XCOFF64")
            # ECOFF:
            # https://web.archive.org/web/20160305114748/http://h41361.www4.hp.com/docs/base_doc/DOCUMENTATION/V50A_ACRO_SUP/OBJSPEC.PDF
            if magic_bytes[:2] in ("\x83\x01", "\x88\x01", "\x8f\x01"):
                filetype_matches.append("ECOFF")
            # AR:
            # https://www.garykessler.net/library/file_sigs.html
            if magic_bytes[:8] == b"!<arch>\n":
                filetype_matches.append("AR_LIB")
            # OMF:
            # https://github.com/file/file/blob/c8bba134ac1f3c9f5/magic/Magdir/msvc#L22
            if (
                int.from_bytes(magic_bytes[0:4], byteorder="big", signed=False) & 0xFF0F80FF
            ) == 0xF00D0000:
                filetype_matches.append("OMF_LIB")
            # U-Boot/uImage
            # https://github.com/u-boot/u-boot/blob/master/include/image.h#L313
            if magic_bytes[:4] == b"\x27\x05\x19\x56":
                filetype_matches.append("UIMAGE")
            # zlib:
            # https://www.rfc-editor.org/rfc/rfc1950
            if len(magic_bytes) >= 2:
                cmf = magic_bytes[0]
                flg = magic_bytes[1]
                cm = cmf & 0x0F
                if cm == 8:
                    if (cmf * 256 + flg) % 31 == 0:
                        filetype_matches.append("ZLIB")
            # cpio:
            # https://commons.apache.org/proper/commons-compress/apidocs/org/apache/commons/compress/archivers/cpio/CpioArchiveEntry.html
            if int.from_bytes(magic_bytes[:2], byteorder="big", signed=False) == 0o70707:
                filetype_matches.append("CPIO_BIN big")
            if int.from_bytes(magic_bytes[:2], byteorder="little", signed=False) == 0o70707:
                filetype_matches.append("CPIO_BIN little")
            if magic_bytes[:6] == b"070707":
                filetype_matches.append("CPIO_ASCII_OLD")
            if magic_bytes[:6] == b"070701":
                filetype_matches.append("CPIO_ASCII_NEW")
            if magic_bytes[:6] == b"070702":
                filetype_matches.append("CPIO_ASCII_NEW_CRC")
            # zstd:
            # https://datatracker.ietf.org/doc/html/rfc8878
            if magic_bytes[:4] == b"\x28\xb5\x2f\xfd":
                filetype_matches.append("ZSTANDARD")
            if magic_bytes[:4] == b"\x37\xa4\x30\xec":
                filetype_matches.append("ZSTANDARD_DICTIONARY")
            # iso:
            # https://en.wikipedia.org/wiki/List_of_file_signatures
            for offset in (0x8001, 0x8801, 0x9001):
                f.seek(offset)
                iso_bytes = f.read(5)
                if iso_bytes == b"CD001":
                    filetype_matches.append("ISO_9660_CD")
            f.seek(0)
            # MacOS dmg:
            # https://en.wikipedia.org/wiki/List_of_file_signatures
            file_size = f.seek(0, os.SEEK_END)
            if file_size >= 512:
                f.seek(-512, os.SEEK_END)
                macos_bytes = f.read(4)
                if macos_bytes[0:4] == b"koly":
                    filetype_matches.append("MACOS_DMG")
            # rpm:
            # https://rpm-software-management.github.io/rpm/manual/
            if magic_bytes[:4] == b"\xed\xab\xee\xdb":
                filetype_matches.append("RPM Package")

    except FileNotFoundError:
        return None

    return filetype_matches if filetype_matches else None



import json
import uuid as uuid_module
from typing import Tuple

import click
import networkx as nx
from loguru import logger

from surfactant.configmanager import ConfigManager
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager
from surfactant.sbomtypes._sbom import SBOM
from surfactant.sbomtypes._system import System


@click.argument("sbom_outfile", envvar="SBOM_OUTPUT", type=click.File("w"), required=True)
@click.argument("input_sboms", type=click.File("r"), required=True, nargs=-1)
@click.option(
    "--config_file",
    type=click.File("r"),
    required=False,
    help="Config file for controlling some aspects of the merged SBOM, primarily the creation of a new top-level system object (settings here will typically take precedence over command line options)",
)
@click.option(
    "--output_format",
    is_flag=False,
    default=ConfigManager().get(
        "core", "output_format", fallback="surfactant.output.cytrics_writer"
    ),
    help="SBOM output format, options=surfactant.output.[cytrics|csv|spdx]_writer",
)
@click.option(
    "--input_format",
    is_flag=False,
    default="surfactant.input_readers.cytrics_reader",
    help="SBOM input format, assumes that all input SBOMs being merged have the same format, options=surfactant.input_readers.[cytrics|cyclonedx|spdx]_reader",
)
@click.option(
    "--system_uuid",
    is_flag=False,
    help="System UUID to use for relationships to a top-level system object",
)
@click.option(
    "--system_relationship",
    is_flag=False,
    default="Contains",
    show_default=True,
    help="Relationship type between merged SBOM contents to a top-level system object",
)
@click.option(
    "--add_system/--no_add_system",
    default=False,
    show_default=True,
    help="Create a top-level system entry for tying together the merged SBOM components. When disabled, relationships will still be created to a provided system UUID",
)
@click.command("merge")
# pylint: disable-next=too-many-positional-arguments
def merge_command(
    input_sboms,
    sbom_outfile,
    config_file,
    output_format,
    input_format,
    system_uuid,
    system_relationship,
    add_system,
):
    """Merge two or more INPUT_SBOMS together into SBOM_OUTFILE.

    An optional CONFIG_FILE can be supplied to specify a root system entry
    """
    pm = get_plugin_manager()
    output_writer = find_io_plugin(pm, output_format, "write_sbom")
    input_reader = find_io_plugin(pm, input_format, "read_sbom")

    sboms = []
    for sbom in input_sboms:
        sboms.append(input_reader.read_sbom(sbom))
    config = None
    if config_file:
        config = json.load(config_file)
    merge(sboms, sbom_outfile, config, output_writer, system_uuid, system_relationship, add_system)


# pylint: disable-next=too-many-positional-arguments
def merge(
    input_sboms,
    sbom_outfile,
    config,
    output_writer,
    system_uuid=None,
    system_relationship="Contains",
    add_system=False,
):
    """Merge two or more SBOMs, then optionally wrap in a top‐level system."""
    # Merge all input SBOMs into the first one
    merged_sbom = input_sboms[0]
    for sbom_m in input_sboms[1:]:
        merged_sbom.merge(sbom_m)

    # Find root nodes: those with zero incoming edges
    roots = [n for n, deg in merged_sbom.graph.in_degree() if deg == 0]
    logger.info(f"ROOT NODES: {roots}")

    # Detect any directed cycles
    cycles = list(nx.simple_cycles(merged_sbom.graph))
    if cycles:
        logger.warning(f"SBOM CYCLE(S) DETECTED: {cycles}")
    else:
        logger.info("No cycles detected in SBOM graph")

    # Prepare (or suppress) the top‐level system entry
    if config and "system" in config and "UUID" in config["system"]:
        if any(s.UUID == config["system"]["UUID"] for s in merged_sbom.systems):
            add_system = False

    system_obj, using_random = create_system_object(merged_sbom, config, system_uuid)
    if add_system:
        merged_sbom.systems.append(system_obj)

    # Attach a system‐to‐root relationship for each root
    if not using_random or add_system:
        if config and "systemRelationship" in config:
            system_relationship = config["systemRelationship"]
        for root_uuid in roots:
            merged_sbom.create_relationship(system_obj.UUID, root_uuid, system_relationship)
    else:
        logger.warning(
            "No top‐level system relationships added; "
            "either specify --add_system or provide a system UUID."
        )

    # Write out
    output_writer.write_sbom(merged_sbom, sbom_outfile)


def create_system_object(sbom: SBOM, config=None, system_uuid=None) -> Tuple[System, bool]:
    """Function to create an accurate system object

    Positional arguments:
        sbom (SBOM): The SBOM the system object is being created for.
        config: The user specified config json (Optional).

    Returns:
        Tuple[System, bool]: The created system object and a boolean indicating if a random UUID was used.
    """

    system = {}
    using_random_uuid = False
    if config and "system" in config:
        system = config["system"]

    # system_uuid supplied via command line overrides config file UUID
    if system_uuid:
        system["UUID"] = system_uuid
    elif "UUID" not in system:
        # No UUID, generate a random one...
        using_random_uuid = True
        system["UUID"] = str(uuid_module.uuid4())
    # check if the UUID appears valid based on the CyTRICS schema
    elif not sbom.is_valid_uuid4(system["UUID"]):
        invalid_uuid = system["UUID"]
        logger.error(f"Invalid uuid4 given ({invalid_uuid}) for the system")

    if "name" not in system:
        system["name"] = ""
    captureStart = -1
    captureEnd = -1
    for sw in sbom.software:
        if captureStart and sw.captureTime:
            if captureStart == -1 or sw.captureTime < captureStart:
                captureStart = sw.captureTime
        if captureEnd and sw.captureTime:
            if captureEnd == -1 or sw.captureTime > captureEnd:
                captureEnd = sw.captureTime
    if "captureStart" not in system or not system["captureStart"]:
        system["captureStart"] = captureStart
    if "captureEnd" not in system or not system["captureEnd"]:
        system["captureEnd"] = captureEnd
    return System(**system), using_random_uuid



import gzip
import importlib.util
import json
import logging
import os
import shutil
import subprocess
import sys

import docker
import pytest

from surfactant.configmanager import ConfigManager

logging.basicConfig(level=logging.INFO)

# Globals
PLUGIN_NAME = "surfactantplugin_grype"
DOCKER_IMAGE = "hello-world"
GRYPE_OUTPUT_KEY = "grype_output"


def run_command(command: str) -> str:
    """Run a shell command and return its output."""
    result = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)
    return result.stdout.strip()


@pytest.fixture(scope="session")
def setup_environment_fixture():
    """Fixture to install Grype dependency."""
    check_dependency_availability()

    # Verify and install required tools
    install_grype()


def check_dependency_availability():
    # Check for Docker SDK
    if importlib.util.find_spec("docker") is None:
        pytest.skip("Docker SDK for Python is not installed.")

    # Check for Surfactant
    if importlib.util.find_spec("surfactant") is None:
        pytest.skip("Surfactant is not installed.")


def install_grype() -> None:
    """Install Grype if not already installed."""
    try:
        # Try to use an existing Grype installation
        output = subprocess.run(
            ["grype", "--version"], capture_output=True, text=True, check=True
        ).stdout.strip()
        logging.info("Grype is already installed: '%s'", output)
    except (subprocess.CalledProcessError, FileNotFoundError):
        logging.info("Installing Grype...")
        # Use the official installer script but with subprocess list syntax instead of shell=True
        subprocess.run(
            [
                "curl",
                "-sSfL",
                "https://raw.githubusercontent.com/anchore/grype/main/install.sh",
                "-o",
                "grype-installer.sh",
            ],
            check=True,
        )

        # Make the installer executable
        os.chmod("grype-installer.sh", 0o755)

        # Run the installer
        subprocess.run(["./grype-installer.sh", "-b", "/usr/local/bin"], check=True)

        # Clean up
        os.remove("grype-installer.sh")

        logging.info("Grype installed successfully.")


def enable_plugin(plugin_name):
    """Enable a plugin using the Surfactant API directly."""
    # Get config manager
    config_manager = ConfigManager()

    # Retrieve the current list of disabled plugins
    section = "core"
    section_key = "disable_plugins"
    disabled_plugins = config_manager.get(section, section_key, [])

    # If the plugin is in the disabled list, remove it
    if plugin_name in disabled_plugins:
        disabled_plugins.remove(plugin_name)
        config_manager.set(section, section_key, disabled_plugins)

    # Verify the plugin is now enabled (not in the disabled list)
    disabled_plugins = config_manager.get(section, section_key, [])
    assert plugin_name not in disabled_plugins, f"{plugin_name} still found in disabled plugins"

    logging.info("Plugin '%s' is now enabled", plugin_name)


def disable_plugin(plugin_name):
    """Disable a plugin using the Surfactant API directly."""
    # Get config manager
    config_manager = ConfigManager()

    # Retrieve the current list of disabled plugins
    section = "core"
    section_key = "disable_plugins"
    disabled_plugins = config_manager.get(section, section_key, [])

    # If the plugin is not in the disabled list, add it
    if plugin_name not in disabled_plugins:
        disabled_plugins.append(plugin_name)
        config_manager.set(section, section_key, disabled_plugins)

    # Verify the plugin is now disabled (in the disabled list)
    disabled_plugins = config_manager.get(section, section_key, [])
    assert plugin_name in disabled_plugins, f"{plugin_name} not found in disabled plugins"

    logging.info("Plugin '%s' is now disabled", plugin_name)


def run_surfactant_generate(config_file, output_sbom_path):
    """Run surfactant generate using Python subprocess in a platform-independent way."""
    # Use the Python executable from the current environment
    python_exe = sys.executable

    # Build the command using list form to avoid shell injection issues
    cmd = [python_exe, "-m", "surfactant", "generate", config_file, str(output_sbom_path)]

    # Run the command
    logging.info("Running command: %s", " ".join(cmd))
    result = subprocess.run(cmd, check=True, capture_output=True, text=True)

    logging.info("Successfully generated SBOM: '%s'", output_sbom_path)
    return result.stdout


@pytest.fixture(scope="session", name="config_and_tarball_fixture")
def create_config_and_tarball_fixture(tmp_path_factory):
    """
    Fixture to create the configuration file and Docker tarball for testing.
    The tarball contains the 'hello-world' Docker container filesystem.
    """
    # Create a session-scoped temporary directory
    temp_dir = tmp_path_factory.mktemp("session_temp")

    # Initialize Docker client
    logging.info("Initializing Docker client...")
    docker_client = docker.from_env()

    # Pull Docker image
    logging.info("Pulling Docker image: '%s':latest", DOCKER_IMAGE)
    docker_client.images.pull(DOCKER_IMAGE, tag="latest")
    logging.info("Successfully pulled Docker image: '%s':latest", DOCKER_IMAGE)

    # Save Docker image to tar file
    temp_tar_file = temp_dir / "myimage_latest.tar"
    logging.info("Saving Docker image to file: '%s'", temp_tar_file)
    with open(temp_tar_file, "wb") as f:
        bytes_written = 0
        for chunk in docker_client.images.get(f"{DOCKER_IMAGE}:latest").save(named=True):
            f.write(chunk)
            bytes_written += len(chunk)
    logging.info(
        "Successfully saved Docker image to file: '%s' ('%s' bytes)", temp_tar_file, bytes_written
    )

    # Change ownership of the file to the current user
    os.chmod(temp_tar_file, 0o644)

    # Export the container's filesystem to a tarball
    tarball_file = temp_dir / "myimage_latest.tar.gz"

    # Compress the docker image and save to file
    with open(temp_tar_file, "rb") as f_in:
        logging.info("Compressing Docker image tar file with gzip...")
        with gzip.open(tarball_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    # Remove the temporary tar file
    try:
        os.remove(temp_tar_file)
        logging.info("Temporary tar file removed")
    except OSError as e:
        logging.warning("Failed to remove temporary tar file: '%s'", e)

    # Remove Docker image
    logging.info("Removing Docker image: '%s':latest", DOCKER_IMAGE)
    docker_client.images.remove(f"{DOCKER_IMAGE}:latest")
    logging.info("Successfully removed Docker image: '%s':latest", DOCKER_IMAGE)

    # Create the configuration file
    config_data = [{"extractPaths": [str(tarball_file)], "installPrefix": "/usr/"}]
    config_file = temp_dir / "config_dockertball.json"
    with open(config_file, "w", encoding="utf-8") as f:
        json.dump(config_data, f, indent=4)
    logging.info("Configuration file created: '%s'", config_file)

    return str(config_file), str(tarball_file)


def test_debug_create_config_and_tarball(config_and_tarball_fixture):
    # pytest test_grype.py -k test_debug_create_config_and_tarball -v
    # Call the fixture and unpack its return values
    config_file, tarball_file = config_and_tarball_fixture

    # Log the outputs for debugging
    logging.info("Config file path: '%s'", config_file)
    logging.info("Tarball file path: '%s'", tarball_file)

    # Assert that the files were created successfully
    assert config_file is not None, "Config file was not created"
    assert tarball_file is not None, "Tarball file was not created"


@pytest.mark.usefixtures("setup_environment_fixture", "config_and_tarball_fixture")
def test_surfactant_generate(config_and_tarball_fixture, tmp_path_factory):
    """Test the Surfactant generate command with the Grype plugin."""
    # Unpack the fixture values
    config_file, tarball_file = config_and_tarball_fixture

    # Create a temporary directory for the test
    temp_dir = tmp_path_factory.mktemp("test_temp")

    # **********************
    # **** Enabled Test ***
    # **********************

    # Enable and verify the Grype plugin is enabled
    enable_plugin(PLUGIN_NAME)

    # Run the Surfactant generate command (with Grype enabled)
    output_enabled_sbom = temp_dir / "docker_tball_grype-enabled_sbom.json"
    logging.info(config_file)
    with open(config_file, "r", encoding="utf-8") as f:
        config_out = json.load(f)
    logging.info(json.dumps(config_out, indent=4))
    logging.info("Running surfactant generate via API for: '%s'", output_enabled_sbom)
    run_surfactant_generate(config_file, output_enabled_sbom)

    # Verify the SBOM file is created
    assert output_enabled_sbom.exists(), f"SBOM file not created: {output_enabled_sbom}"

    # Read and parse the SBOM
    with open(output_enabled_sbom, "r", encoding="utf-8") as f:
        sbom_enabled = json.load(f)

    # Assert that the Grype output is present
    logging.info("ENABLED")
    logging.info(json.dumps(sbom_enabled, indent=4))
    assert any(GRYPE_OUTPUT_KEY in entry for entry in sbom_enabled["software"][0]["metadata"]), (
        "Grype output should be present when the plugin is enabled"
    )

    # Assert that the Grype output is empty (in this specific test case)
    assert all(
        entry.get(GRYPE_OUTPUT_KEY) == []
        for entry in sbom_enabled["software"][0]["metadata"]
        if GRYPE_OUTPUT_KEY in entry
    ), "Grype output should be empty for a minimal tarball with no vulnerabilities"

    # **********************
    # **** Disabled Test ***
    # **********************

    # Disable and verify the Grype plugin is disabled
    disable_plugin(PLUGIN_NAME)

    # Disable the Grype plugin and verify the plugin is disabled
    output_disabled_sbom = temp_dir / "docker_tball_grype-disabled_sbom.json"
    logging.info("Running surfactant generate via API for: '%s'", output_disabled_sbom)
    run_surfactant_generate(config_file, output_disabled_sbom)

    # Verify the SBOM file is created
    assert output_disabled_sbom.exists(), f"SBOM file not created: {output_disabled_sbom}"

    # Read and parse the SBOM
    with open(output_disabled_sbom, "r", encoding="utf-8") as f:
        sbom_disabled = json.load(f)

    # Assert that the Grype output is not present
    logging.info("DISABLED")
    logging.info(json.dumps(sbom_disabled, indent=4))
    assert not any(
        GRYPE_OUTPUT_KEY in entry for entry in sbom_disabled["software"][0]["metadata"]
    ), "Grype output should not be present when the plugin is disabled"

    # ************************
    # *** Test consistency ***
    # ************************

    # Compare the two SBOMs for consistency (except for Grype output)
    assert sbom_disabled["software"][0]["fileName"] == sbom_enabled["software"][0]["fileName"], (
        "File names should match between disabled and enabled cases"
    )
    assert sbom_disabled["software"][0]["sha256"] == sbom_enabled["software"][0]["sha256"], (
        "SHA256 hashes should match between disabled and enabled cases"
    )



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# Extracts information on U-Boot/uImage files
# https://github.com/u-boot/u-boot/blob/master/include/image.h
# https://github.com/u-boot/u-boot/blob/master/boot/image.c

import struct
from typing import List, Tuple

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software

# Compression types is not a tuple because (other than none), the id, short name, and message more or less match
COMPRESSION_TYPES = {
    0: "None",
    1: "gzip",
    2: "bzip2",
    3: "lzma",
    4: "lzo",
    5: "lz4",
    6: "zstd",
}

# Operating System Code (IH_OS_id, short name, long name/description)
OPERATING_SYSTEMS = {
    0: ("INVALID", "invalid", "Invalid OS"),
    1: ("OPENBSD", "openbsd", "OpenBSD"),
    2: ("NETBSD", "netbsd", "NetBSD"),
    3: ("FREEBSD", "freebsd", "FreeBSD"),
    4: ("4_4BSD", "4_4bsd", "4_4BSD"),  # not sure if long description should be 4.4BSD
    5: ("LINUX", "linux", "Linux"),
    6: ("SVR4", "svr4", "SVR4"),
    7: ("ESIX", "esix", "Esix"),
    8: ("SOLARIS", "solaris", "Solaris"),
    9: ("IRIX", "irix", "Irix"),
    10: ("SCO", "sco", "SCO"),
    11: ("DELL", "dell", "Dell"),
    12: ("NCR", "ncr", "NCR"),
    13: ("LYNXOS", "lynxos", "LynxOS"),  # short/long name mapping table is missing this one
    14: ("VXWORKS", "vxworks", "VxWorks"),
    15: ("PSOS", "psos", "pSOS"),
    16: ("QNX", "qnx", "QNX"),
    17: ("U_BOOT", "u-boot", "U-Boot"),
    18: ("RTEMS", "rtems", "RTEMS"),
    19: ("ARTOS", "artos", "ARTOS"),  # short/long name mapping table is missing this one
    20: ("UNITY", "unity", "Unity OS"),  # short/long name mapping table is missing this one
    21: ("INTEGRITY", "integrity", "INTEGRITY"),
    22: ("OSE", "ose", "Enea OSE"),
    23: ("PLAN9", "plan9", "Plan 9"),
    24: ("OPENRTOS", "openrtos", "OpenRTOS"),
    25: ("ARM_TRUSTED_FIRMWARE", "arm-trusted-firmware", "ARM Trusted Firmware"),
    26: ("TEE", "tee", "Trusted Execution Environment"),
    27: ("OPENSBI", "opensbi", "RISC-V OpenSBI"),
    28: ("EFI", "efi", "EFI Firmware"),
    29: ("ELF", "elf", "ELF Image"),
}

# CPU Architecture Code (supported by Linux) tuples (IH_ARCH_id, short name, long name/description)
ARCHITECTURES = {
    0: ("INVALID", "invalid", "Invalid ARCH"),
    1: ("ALPHA", "alpha", "Alpha"),
    2: ("ARM", "arm", "ARM"),
    3: ("I386", "x86", "Intel x86"),
    4: ("IA64", "ia64", "IA64"),
    5: ("MIPS", "mips", "MIPS"),
    6: ("MIPS64", "mips64", "MIPS 64 Bit"),
    7: ("PPC", "powerpc", "PowerPC"),  # also has an alias short name of ppc
    8: ("S390", "s390", "IBM S390"),
    9: ("SH", "sh", "SuperH"),
    10: ("SPARC", "sparc", "SPARC"),
    11: ("SPARC64", "sparc64", "SPARC 64 Bit"),
    12: ("M68K", "m68k", "M68K"),
    13: ("NIOS", "nios", "Nios-32"),  # short/long name mapping table is missing this one
    14: ("MICROBLAZE", "microblaze", "MicroBlaze"),
    15: ("NIOS2", "nios2", "NIOS II"),
    16: ("BLACKFIN", "blackfin", "Blackfin"),
    17: ("AVR32", "avr32", "AVR32"),
    18: (
        "ST200",
        "st200",
        "STMicroelectronics ST200",
    ),  # short/long name mapping table is missing this one
    19: ("SANDBOX", "sandbox", "Sandbox"),  # test only
    20: ("NDS32", "nds32", "NDS32"),
    21: ("OPENRISC", "or1k", "OpenRISC 1000"),
    22: ("ARM64", "arm64", "AArch64"),
    23: ("ARC", "arc", "ARC"),
    24: ("X86_64", "x86_64", "AMD x86_64"),
    25: ("XTENSA", "xtensa", "Xtensa"),
    26: ("RISCV", "riscv", "RISC-V"),
}

# Image type tuples (IH_TYPE_id, short name, long name/description)
# Summary of Image Types:
#  - Standalone Programs: Directly runnable in the environment provided by U-Boot; can return control to U-Boot.
#  - OS Kernel Images: Images of embedded OS that take over control completely; cannot re-enter U-Boot except by resetting the CPU.
#  - RAMDisk Images: Data blocks whose parameters are passed to an OS kernel being started.
#  - Multi-File Images: Contain several images (e.g., OS kernel and RAMDisks); useful for booting over the network.
#  - Firmware Images: Binary images containing firmware (e.g., U-Boot or FPGA images) usually programmed to flash memory.
#  - Script Files: Command sequences executed by U-Boot's command interpreter; useful for shell scripts.
IMAGE_TYPES = {
    0: ("INVALID", "invalid", "Invalid Image"),
    1: ("STANDALONE", "standalone", "Standalone Program"),
    2: ("KERNEL", "kernel", "Kernel Image"),
    3: ("RAMDISK", "ramdisk", "RAMDisk Image"),
    4: ("MULTI", "multi", "Multi-File Image"),
    5: ("FIRMWARE", "firmware", "Firmware"),
    6: ("SCRIPT", "script", "Script"),
    7: ("FILESYSTEM", "filesystem", "Filesystem Image"),
    8: ("FLATDT", "flat_dt", "Flat Device Tree"),
    9: ("KWBIMAGE", "kwbimage", "Kirkwood Boot Image"),
    10: ("IMXIMAGE", "imximage", "Freescale i.MX Boot Image"),
    11: ("UBLIMAGE", "ublimage", "Davinci UBL Image"),
    12: ("OMAPIMAGE", "omapimage", "TI OMAP SPL with GP CH"),
    13: ("AISIMAGE", "aisimage", "Davinci AIS Image"),
    14: ("KERNEL_NOLOAD", "kernel_noload", "Kernel Image (no loading done)"),
    15: ("PBLIMAGE", "pblimage", "Freescale PBL Boot Image"),
    16: ("MXSIMAGE", "mxsimage", "Freescale MXS Boot Image"),
    17: ("GPIMAGE", "gpimage", "TI Keystone SPL Image"),
    18: ("ATMELIMAGE", "atmelimage", "ATMEL ROM-Boot Image"),
    19: ("SOCFPGAIMAGE", "socfpgaimage", "Altera SOCFPGA CV/AV preloader"),
    20: ("X86_SETUP", "x86_setup", "x86 setup.bin"),
    21: (
        "LPC32XXIMAGE",
        "lpc32xximage",
        "LPC32XX Boot Image",
    ),  # image.h for u-boot has incorrect comment on this line
    22: ("LOADABLE", "", "A list of typeless images"),
    23: ("RKIMAGE", "rkimage", "Rockchip Boot Image"),
    24: ("RKSD", "rksd", "Rockchip SD Boot Image"),
    25: ("RKSPI", "rkspi", "Rockchip SPI Boot Image"),
    26: ("ZYNQIMAGE", "zynqimage", "Xilinx Zynq Boot Image"),
    27: ("ZYNQMPIMAGE", "zynqmpimage", "Xilinx ZynqMP Boot Image"),
    28: ("ZYNQMPBIF", "zynqmpbif", "Xilinx ZynqMP Boot Image (bif)"),
    29: ("FPGA", "fpga", "FPGA Image"),
    30: ("VYBRIDIMAGE", "vybridimage", "Vybrid Boot Image"),  # .vyb
    31: ("TEE", "tee", "Trusted Execution Environment Image"),
    32: ("FIRMWARE_IVT", "firmware_ivt", "Firmware with HABv4 IVT"),
    33: ("PMMC", "pmmc", "TI Power Management Micro-Controller Firmware"),
    34: ("STM32IMAGE", "stm32image", "STMicroelectronics STM32 Image"),
    35: ("SOCFPGAIMAGE_V1", "socfpgaimage_v1", "Altera SOCFPGA A10 preloader"),
    36: ("MTKIMAGE", "mtk_image", "MediaTek BootROM loadable Image"),
    37: ("IMX8MIMAGE", "imx8mimage", "NXP i.MX8M Boot Image"),
    38: ("IMX8IMAGE", "imx8image", "NXP i.MX8 Boot Image"),
    39: ("COPRO", "copro", "Coprocessor Image"),
    40: ("SUNXI_EGON", "sunxi_egon", "Allwinner eGON Boot Image"),
    41: ("SUNXI_TOC0", "sunxi_toc0", "Allwinner TOC0 Boot Image"),
    42: ("FDT_LEGACY", "fdt_legacy", "Legacy Image with Flat Device Tree"),
    43: ("RENESAS_SPKG", "spkgimage", "Renesas SPKG Image"),
    44: ("STARFIVE_SPL", "sfspl", "StarFive SPL Image"),
    45: ("TFA_BL31", "tfa-bl31", "TFA BL31 Image"),
}

# Phase that the image is intended for
# Combined with image type for composite type (but appears to not be used in legacy u-boot header)
IMAGE_PHASES = {
    0: ("NONE", "none", "any"),
    1: ("U_BOOT", "u-boot", "U-Boot phase"),
    2: ("SPL", "spl", "SPL Phase"),
}


def _parse_uimage_header(fname: str) -> dict:
    # Image header is 64 bytes
    # struct layout (big-endian):
    # magic(4), header_crc(4), timestamp(4), size(4), load(4), ep(4), data_crc(4),
    # os(1), arch(1), im_type(1), comp_type(1), name(32)
    with open(fname, "rb") as f:
        data = f.read(64)
    try:
        (
            _magic,
            header_crc,
            timestamp,
            data_size,
            load_addr,
            entry_point,
            data_crc,
            os_type,
            arch,
            image_type,
            compression_type,
        ) = struct.unpack(">IIIIIIIBBBB", data[:32])
    except struct.error as e:
        logger.warning(f"Error unpacking uImage header in file {fname}")
        raise ValueError(f"Error unpacking uImage header in file {fname}") from e
    image_name = data[32:64].rstrip(b"\x00").decode("ascii", errors="replace")

    # for multi (or script?) type images, go to data right after header; count non-zero array uint32_t sizes to get number of parts
    # image data starts at start_data + (count+1)*sizeof(uint32_t) to account for last null entry
    # sizes are stored in the file as big endian, and are the size for the given component
    # to find data offset for a particular component index, add up each preceeding component size rounded up to 4-bytes: offset += (uimage_to_cpu(size[i]) + 3) & ~3 ;

    # Related image types are fit/fdt, and Android... need to investigate if they will have the legacy u-boot header or not
    # could be that fdt image header is nested in data section only when image type is FDT_LEGACY
    # fit/fdt images, may be possible to check for fdt header magic 0xd00dfeed
    # Android boot magic is "ANDROID!"
    return {
        "header_crc": hex(header_crc),
        "timestamp": timestamp,
        "data_size": data_size,
        "load_addr": hex(load_addr),
        "entry_point": hex(entry_point),
        "data_crc": hex(data_crc),
        "os": OPERATING_SYSTEMS.get(os_type, (str(os_type), "", ""))[0],
        "os_description": OPERATING_SYSTEMS.get(os_type, ("", "", "Unknown OS"))[2],
        "arch": ARCHITECTURES.get(arch, (str(arch), "", ""))[0],
        "arch_description": ARCHITECTURES.get(arch, ("", "", "Unknown Architecture"))[2],
        "image_type": IMAGE_TYPES.get(image_type, (str(image_type), "", ""))[0],
        "image_type_description": IMAGE_TYPES.get(image_type, ("", "", "Unknown Image"))[2],
        "compression_type": COMPRESSION_TYPES.get(compression_type, str(compression_type)),
        "name": image_name,
    }


def supports_file(filetype: List[str]) -> bool:
    return "UIMAGE" in filetype


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: List[str],
    software_field_hints: List[Tuple[str, object, int]],
) -> object:
    if not supports_file(filetype):
        return None
    try:
        uimage_header = _parse_uimage_header(filename)
        if "name" in uimage_header:
            software_field_hints.append(("name", uimage_header["name"], 40))
        return {"uimage_header": uimage_header}
    except ValueError as e:
        return None



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.utils.ahocorasick import AhoCorasick, build_regex_literal_matcher


class TestAhoCorasick:
    def test_initialization(self):
        """Test initialization of the AhoCorasick automaton."""
        ac = AhoCorasick()
        assert not ac.is_bytes
        assert ac.encoding == "utf-8"
        assert not ac.built
        assert not ac.pattern_prefixes

        ac_bytes = AhoCorasick(is_bytes=True, encoding="latin-1")
        assert ac_bytes.is_bytes
        assert ac_bytes.encoding == "latin-1"

    def test_add_pattern(self):
        """Test adding patterns to the automaton."""
        ac = AhoCorasick()
        ac.add_pattern("hello", 1, "hello")

        # Check that the pattern was added correctly
        node = ac.root
        for char in "hello":
            assert char in node.goto
            node = node.goto[char]
        assert node.out == [1]
        assert ac.pattern_prefixes == {1: "hello"}

    def test_add_pattern_bytes(self):
        """Test adding patterns with byte handling."""
        ac = AhoCorasick(is_bytes=True)
        ac.add_pattern(b"hello", 1, "hello")
        ac.add_pattern("world", 2, "world")  # Should be converted to bytes

        # Check pattern addition with bytes
        node = ac.root
        for byte in b"hello":
            assert byte in node.goto
            node = node.goto[byte]
        assert node.out == [1]

        # Check string converted to bytes
        node = ac.root
        for byte in b"world":
            assert byte in node.goto
            node = node.goto[byte]
        assert node.out == [2]

    def test_build_and_search_single_pattern(self):
        """Test building and searching with a single pattern."""
        ac = AhoCorasick()
        ac.add_pattern("hello", 1, "hello")
        ac.build_automaton()

        assert ac.built

        # Test exact match
        results = ac.search("hello world")
        assert results == {1: [0]}

        # Test no match
        results = ac.search("hi there")
        assert not results

        # Test multiple occurrences
        results = ac.search("hello hello hello")
        assert results == {1: [0, 6, 12]}

    def test_search_multiple_patterns(self):
        """Test searching with multiple patterns."""
        ac = AhoCorasick()
        ac.add_pattern("he", 1, "he")
        ac.add_pattern("she", 2, "she")
        ac.add_pattern("his", 3, "his")
        ac.add_pattern("hers", 4, "hers")
        ac.build_automaton()

        results = ac.search("she said he told his and hers story")
        assert results[1] == [1, 9, 25]  # "he"
        assert results[2] == [0]  # "she"
        assert results[3] == [17]  # "his"
        assert results[4] == [25]  # "hers"

    def test_overlapping_patterns(self):
        """Test handling of overlapping patterns."""
        ac = AhoCorasick()
        ac.add_pattern("abcd", 1, "abcd")
        ac.add_pattern("bcd", 2, "bcd")
        ac.add_pattern("cd", 3, "cd")
        ac.build_automaton()

        results = ac.search("abcd")
        assert set(results.keys()) == {1, 2, 3}
        assert results[1] == [0]  # "abcd"
        assert results[2] == [1]  # "bcd"
        assert results[3] == [2]  # "cd"

    def test_bytes_search(self):
        """Test searching with bytes."""
        ac = AhoCorasick(is_bytes=True)
        ac.add_pattern(b"hello", 1, "hello")
        ac.build_automaton()

        # Test with bytes input
        results = ac.search(b"hello world")
        assert results == {1: [0]}

        # Test with string input that gets converted
        results = ac.search("hello world")
        assert results == {1: [0]}

    def test_empty_and_edge_cases(self):
        """Test edge cases."""
        ac = AhoCorasick()

        # Empty pattern should be ignored in build_regex_prefix_matcher
        # but let's test directly
        ac.add_pattern("", 1, "")
        ac.build_automaton()
        results = ac.search("any text")
        # Empty pattern may match at every position or not at all depending on implementation

        # Empty search text
        results = ac.search("")
        assert not results

        # Reset and test with non-empty pattern
        ac = AhoCorasick()
        ac.add_pattern("test", 1, "test")
        ac.build_automaton()
        results = ac.search("")
        assert not results

    def test_build_regex_prefix_matcher(self):
        """Test the build_regex_prefix_matcher function."""
        # Simple literal patterns
        patterns_dict = {1: "hello", 2: "world"}

        ac = build_regex_literal_matcher(patterns_dict)

        # Both patterns should be added as-is
        results = ac.search("hello world")
        assert set(results.keys()) == {1, 2}
        assert results[1] == [0]
        assert results[2] == [6]

        # Test with regex patterns
        regex_patterns = {
            1: "hel+lo",  # Should extract "hel" as prefix
            2: "wo(rld|w)",  # Should extract "wo" as prefix
        }

        ac = build_regex_literal_matcher(regex_patterns)

        # Should match both patterns
        results = ac.search("hello world")
        assert set(results.keys()) == {1, 2}

        # Complex regex with no fixed prefix should be skipped
        regex_patterns = {
            1: "hel+lo",  # Should extract "hel" as prefix
            2: ".*(rld|w)",  # No fixed prefix
        }

        ac = build_regex_literal_matcher(regex_patterns)

        # Should match only pattern 1
        results = ac.search("hello world")
        assert set(results.keys()) == {1}


# Some additional tests, at times testing a slighly higher level of abstraction than the above tests
def test_add_pattern_and_search():
    ac = AhoCorasick()
    ac.add_pattern("hello", 1, "hello")
    ac.add_pattern("world", 2, "world")
    ac.build_automaton()

    text = "hello world, hello again"
    results = ac.search(text)

    assert results[1] == [0, 13], f"Expected [0, 13], but got {results[1]}"
    assert results[2] == [6], f"Expected [6], but got {results[2]}"


def test_add_pattern_with_bytes():
    ac = AhoCorasick(is_bytes=True)
    ac.add_pattern(b"hello", 1, "hello")
    ac.add_pattern(b"world", 2, "world")
    ac.build_automaton()

    text = b"hello world, hello again"
    results = ac.search(text)

    assert results[1] == [0, 13], f"Expected [0, 13], but got {results[1]}"
    assert results[2] == [6], f"Expected [6], but got {results[2]}"


def test_build_regex_literal_matcher():
    patterns = {
        1: r"hello",
        2: r"world",
        3: r"^foo|bar",
    }
    ac = build_regex_literal_matcher(patterns, is_literal=True)

    text = "hello world, foo and bar"
    results = ac.search(text)

    assert results[1] == [0], f"Expected [0], but got {results[1]}"
    assert results[2] == [6], f"Expected [6], but got {results[2]}"
    assert 3 not in results, f"Expected no results for 3 but got {results}"


def test_no_match():
    ac = AhoCorasick()
    ac.add_pattern("test", 1, "test")
    ac.build_automaton()

    text = "no matches here"
    results = ac.search(text)

    assert not results, f"Expected an empty dictionary, but got {results}"


def test_prefix_handling():
    ac = AhoCorasick()
    ac.add_pattern("prefix", 1, "prefix")
    ac.add_pattern("prefix123", 2, "prefix123")
    ac.build_automaton()

    text = "prefix123 and prefix"
    results = ac.search(text)

    assert results[1] == [0, 14], f"Expected [0, 14], but got {results[1]}"
    assert results[2] == [0], f"Expected [0], but got {results[2]}"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import pathlib

from surfactant.filetypeid.id_magic import identify_file_type
from surfactant.infoextractors.uimage_file import extract_file_info
from surfactant.sbomtypes import SBOM, Software

# Files are relative to tests/data/a_out_files
_uimage_files = {
    "hello_world.img": {
        "header_crc": "0x92c6061b",
        "timestamp": 1741809003,
        "data_size": 61,
        "load_addr": "0x1234",
        "entry_point": "0x5678",
        "data_crc": "0x21ae6b9e",
        "os": "LINUX",
        "os_description": "Linux",
        "arch": "ARM",
        "arch_description": "ARM",
        "image_type": "FIRMWARE",
        "image_type_description": "Firmware",
        "compression_type": "None",
        "name": "Test uImage",
    },
}


def test_uimage_files():
    base_path = pathlib.Path(__file__).parent.absolute()
    data_dir = os.path.join(base_path, "..", "data", "uimage_files")
    for file_name, expected_values in _uimage_files.items():
        file_path = os.path.join(data_dir, file_name)
        file_type = identify_file_type(file_path)
        assert file_type == ["UIMAGE"]
        sw_field_hints = []
        file_info = extract_file_info(SBOM(), Software(), file_path, file_type, sw_field_hints)
        assert sw_field_hints == [("name", "Test uImage", 40)]
        assert "uimage_header" in file_info
        uimage_header = file_info["uimage_header"]
        for key, expected_value in expected_values.items():
            assert key in uimage_header, f"Missing key: {key}"
            assert uimage_header[key] == expected_value, (
                f"Value mismatch for {key}: expected {expected_value}, got {uimage_header[key]}"
            )


def test_bad_uimage_file():
    base_path = pathlib.Path(__file__).parent.absolute()
    file_path = os.path.join(base_path, "..", "data", "uimage_files", "bad1.img")
    file_type = identify_file_type(file_path)
    assert file_type == ["UIMAGE"]
    sw_field_hints = []
    file_info = extract_file_info(SBOM(), Software(), file_path, file_type, sw_field_hints)
    assert file_info is None



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import os
import pathlib
from typing import Optional

import click
import textual.app
import textual.containers
import textual.css.query
import textual.events
import textual.screen
import textual.types
import textual.widgets
import textual.widgets.button

import surfactant.cmd.generate
import surfactant.cmd.merge


class YesNoScreen(textual.screen.Screen[bool]):
    """Screen that presents a yes/no question"""

    def __init__(self, question: str) -> None:
        self.question = question
        super().__init__()

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Label(self.question)
        yield textual.widgets.Button("Yes", id="yes", variant="success")
        yield textual.widgets.Button("No", id="no")

    @textual.on(textual.widgets.Button.Pressed, "#yes")
    def handle_yes(self) -> None:
        self.dismiss(True)

    @textual.on(textual.widgets.Button.Pressed, "#no")
    def handle_no(self) -> None:
        self.dismiss(False)


class SelectFileButtons(textual.widgets.Static):
    # pylint: disable=too-few-public-methods
    def __init__(self, allow_folder_selection: bool):
        super().__init__()
        self.allow_folder_selection = allow_folder_selection

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("Go up a directory", id="up_dir")
        if self.allow_folder_selection:
            yield textual.widgets.Button("Select directory", id="select_dir")


class SelectFile(textual.screen.ModalScreen[Optional[textual.widgets.DirectoryTree.FileSelected]]):
    """Pop-up to select a file"""

    def __init__(self, allow_folder_selection: bool, start_path: str):
        super().__init__()
        self.allow_folder_selection = allow_folder_selection
        self.dir_selected = start_path

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.DirectoryTree(self.dir_selected, id="file_dir")
        yield SelectFileButtons(self.allow_folder_selection)

    @textual.on(textual.widgets.Button.Pressed, "#up_dir")
    def handle_up_dir(self):
        tree = self.get_child_by_id("file_dir", textual.widgets.DirectoryTree)
        tree.path = tree.path.parent.resolve()
        tree.reload()

    @textual.on(textual.widgets.Button.Pressed, "#select_dir")
    def handle_select_dir(self):
        self.dismiss(self.dir_selected)

    def on_directory_tree_directory_selected(
        self, path: textual.widgets.DirectoryTree.DirectorySelected
    ) -> None:
        self.dir_selected = path.path.as_posix()

    def on_directory_tree_file_selected(
        self, path: textual.widgets.DirectoryTree.FileSelected
    ) -> None:
        self.dismiss(path.path.as_posix())

    def on_key(self, event: textual.events.Key):
        if event.key == "escape":
            self.dismiss(None)


class FileInput(textual.widgets.Static):
    def __init__(
        self, label: str, allow_folder_selection: bool, file_input: Optional[textual.widgets.Input]
    ):
        super().__init__()
        self.label = label
        self.file_input = file_input
        self.allow_folder_selection = allow_folder_selection
        self.input_path = ""

    def compose(self) -> textual.app.ComposeResult:
        if len(self.input_path) == 0:
            yield textual.widgets.Label(f"{self.label} \\[Click to set]")
        else:
            yield textual.widgets.Label(f"{self.label} {self.input_path}")

    def on_click(self):
        def set_path(path: Optional[pathlib.Path]):
            if path:
                if self.file_input and not os.path.isdir(path):
                    # Set the directory and the file separately
                    self.input_path = os.path.dirname(path)
                    self.file_input.value = os.path.basename(path)
                else:
                    # Just set the path
                    self.input_path = path
                self.query_one(textual.widgets.Label).update(f"{self.label} {self.input_path}")

        base_dir = "./"
        if os.path.isfile(self.input_path):
            base_dir = os.path.dirname(self.input_path)
        self.app.push_screen(SelectFile(self.allow_folder_selection, base_dir), set_path)


# pylint: disable-next=too-many-instance-attributes
class GenerateTab(textual.widgets.Static):
    def __init__(self):
        super().__init__()
        self.specimen_context = FileInput("Specimen context:", True, None)
        self.output_name = textual.widgets.Input(placeholder="Output Filename")
        self.output_dir = FileInput("Output directory:", True, self.output_name)
        self.input_sbom = FileInput("Input SBOM:", False, None)
        self.skip_gather = textual.widgets.Checkbox("Skip Gather")
        self.skip_relationships = textual.widgets.Checkbox("Skip Relationships")
        self.input_format = textual.widgets.Select([("CyTRICS", "CyTRICS")], allow_blank=False)
        self.skip_install_path = textual.widgets.Checkbox("Skip Install Path")
        self.recorded_institution = textual.widgets.Input(placeholder="Recorded Institution")
        self.output_format = textual.widgets.Select(
            [("CyTRICS", "CyTRICS"), ("SPDX", "SPDX"), ("CSV", "CSV")], allow_blank=False
        )

    def compose(self) -> textual.app.ComposeResult:
        yield self.specimen_context
        yield self.output_dir
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Output Filename: "), self.output_name
        )
        yield textual.widgets.Rule()
        yield textual.widgets.Label("Optional options:")
        yield self.input_sbom
        yield self.skip_gather
        yield self.skip_relationships
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Input Format: "), self.input_format
        )
        yield self.skip_install_path
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Recorded Institution: "), self.recorded_institution
        )
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Output Format: "), self.output_format
        )
        yield textual.widgets.Button("Run", id="run")

    @textual.on(textual.widgets.Button.Pressed, "#run")
    def handle_run(self) -> None:
        if len(self.specimen_context.input_path) == 0:
            self.app.notify("No input file selected")
            return
        if len(self.output_dir.input_path) == 0:
            self.app.notify("No output directory selected")
            return
        if len(self.output_name.value) == 0:
            self.app.notify("No output name supplied")
            return
        with self.app.suspend():
            args = [
                self.specimen_context.input_path,
                f"{self.output_dir.input_path}/{self.output_name.value}",
            ]
            if len(self.input_sbom.input_path) > 0:
                args.append(self.input_sbom.input_path)
            if self.skip_gather.value:
                args.append("--skip_gather")
            if self.skip_relationships.value:
                args.append("--skip_relationships")
            if len(self.input_format.value) > 0:
                args.append("--input_format")
                args.append(self.input_format.value)
            if self.skip_install_path.value:
                args.append("--skip_install_path")
            args.append("--input_format")
            args.append(self.input_format.value)
            args.append("--output_format")
            args.append(self.output_format.value)
            # pylint: disable-next=no-value-for-parameter
            surfactant.cmd.generate.sbom(args, standalone_mode=False)
            print("Press enter to continue")
            _ = input()
        self.app.refresh()


class InputPath(textual.widgets.Static):
    def __init__(self, desc="Input file:", allow_dir_selection=False):
        super().__init__()
        self.path_selector = FileInput(desc, allow_dir_selection, None)
        self.active = True

    def compose(self) -> textual.app.ComposeResult:
        yield textual.containers.HorizontalGroup(
            textual.widgets.Button("-", id="remove_path"), self.path_selector
        )

    @textual.on(textual.widgets.Button.Pressed, "#remove_path")
    def remove_path(self):
        self.active = False
        self.remove()


class InputPathsHolder(textual.widgets.Static):
    def __init__(self, prompt="Input file:", allow_dir_selection=False):
        super().__init__()
        self.input_paths: list[InputPath] = []
        self.prompt = prompt
        self.allow_dir_selection = allow_dir_selection

    def compose(self) -> textual.app.ComposeResult:
        for m_path in self.input_paths:
            if m_path.active:
                yield textual.containers.HorizontalGroup(textual.widgets.Label("   "), m_path)
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("   "), textual.widgets.Button("+", id="add_input_path")
        )

    @textual.on(textual.widgets.Button.Pressed, "#add_input_path")
    def add_input_path(self):
        self.input_paths.append(InputPath(self.prompt, self.allow_dir_selection))
        self.mount(self.input_paths[-1], before="#add_input_path")

    def add_path(self, path: str):
        self.input_paths.append(InputPath(self.prompt, self.allow_dir_selection))
        self.input_paths[-1].path_selector.input_path = path


class MergeTab(textual.widgets.Static):
    def __init__(self):
        super().__init__()
        self.merge_paths = InputPathsHolder()
        self.output_name = textual.widgets.Input(placeholder="Output Filename")
        self.output_dir = FileInput("Output directory:", True, self.output_name)
        self.input_format = textual.widgets.Select(
            [("CyTRICS", "CyTRICS"), ("SPDX", "SPDX"), ("CSV", "CSV")], allow_blank=False
        )
        self.output_format = textual.widgets.Select(
            [("CyTRICS", "CyTRICS"), ("SPDX", "SPDX"), ("CSV", "CSV")], allow_blank=False
        )
        self.config_file = FileInput("Config File:", False, None)

    def compose(self) -> textual.app.ComposeResult:
        yield self.merge_paths
        yield self.output_dir
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Output Filename: "), self.output_name
        )
        yield textual.widgets.Rule()
        yield textual.widgets.Label("Optional options:")
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Input Format: "), self.input_format
        )
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Output Format: "), self.output_format
        )
        yield self.config_file
        yield textual.widgets.Button("Run", id="run")

    @textual.on(textual.widgets.Button.Pressed, "#run")
    def handle_run(self):
        args = []
        if len(self.merge_paths.merge_paths) == 0:
            self.app.notify("No inputs given")
            return
        if len(self.output_dir.input_path) == 0:
            self.app.notify("No output directory given")
            return
        if len(self.output_name.value) == 0:
            self.app.notify("No output filename given")
            return
        for m_path in self.merge_paths.merge_paths:
            if m_path.active:
                path = m_path.path_selector.input_path
                if len(path) == 0:
                    self.app.notify("One or more inputs not given")
                    return
                args.append(path)
        with self.app.suspend():
            args.append(f"{self.output_dir.input_path}/{self.output_name.value}")
            args.append("--input_format")
            args.append(self.input_format.value)
            args.append("--output_format")
            args.append(self.output_format.value)
            config = self.config_file.input_path
            if len(config) > 0:
                args.append("--config_file")
                args.append(config)
            # pylint: disable-next=no-value-for-parameter
            surfactant.cmd.merge.merge_command(args, standalone_mode=False)
            print("Press enter to continue")
            _ = input()
        self.app.refresh()


class ContextEntry(textual.widgets.Static):
    def __init__(self, header_num):
        super().__init__()
        self.active = True
        self.border_title = str(header_num)
        self.archive = FileInput("Archive:", False, None)
        self.install_prefix = textual.widgets.Input(placeholder="Install Prefix")
        self.container_prefix = textual.widgets.Input(placeholder="Container Prefix")
        self.extract_paths = InputPathsHolder("[Click to set extract path]", True)

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("Delete this entry", id="delete_entry")
        yield self.archive
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Install Prefix: "), self.install_prefix
        )
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Container Prefix: "), self.container_prefix
        )
        yield textual.widgets.Label("Extract Paths:")
        yield self.extract_paths

    @textual.on(textual.widgets.Button.Pressed, "#delete_entry")
    def delete_entry(self):
        def delete_self(do_it: Optional[bool]):
            if do_it:
                self.remove()
                self.active = False

        self.app.push_screen(
            YesNoScreen("Are you sure you want to delete this entry?"), delete_self
        )


class ContextTab(textual.widgets.Static):
    def __init__(self):
        super().__init__()
        self.context_name = textual.widgets.Input(placeholder="Context filename")
        self.context_input = FileInput("Context file directory:", True, self.context_name)
        self.context_entries: list[ContextEntry] = []
        self.context_count = 1

    def compose(self) -> textual.app.ComposeResult:
        yield self.context_input
        yield textual.containers.HorizontalGroup(
            textual.widgets.Label("Context filename: "), self.context_name
        )
        yield textual.containers.HorizontalGroup(
            textual.widgets.Button("Save", id="save"),
            textual.widgets.Label("   "),
            textual.widgets.Button("Load", id="load"),
        )
        yield textual.widgets.Rule()
        yield from self.context_entries
        yield textual.widgets.Button("+", id="add_context_entry")

    @textual.on(textual.widgets.Button.Pressed, "#add_context_entry")
    def add_context_entry(self):
        self.context_entries.append(ContextEntry(self.context_count))
        self.mount(self.context_entries[-1], before="#add_context_entry")
        self.context_count += 1

    @textual.on(textual.widgets.Button.Pressed, "#save")
    def save(self):
        to_save = []
        for entry in self.context_entries:
            to_save.append({})
            write_to = to_save[-1]
            archive = entry.archive.input_path
            if len(archive) > 0:
                write_to["archive"] = archive
            install_prefix = entry.install_prefix.value
            if len(install_prefix) > 0:
                write_to["installPrefix"] = install_prefix
            write_to["extractPaths"] = []
            for path in entry.extract_paths.input_paths:
                if path.active:
                    write_to["extractPaths"].append(path.path_selector.input_path)
            container_prefix = entry.container_prefix.value
            if len(container_prefix) > 0:
                write_to["containerPrefix"] = container_prefix
        file_to_save = self.context_input.input_path + "/" + self.context_name.value
        try:
            with open(file_to_save, "w") as f:
                f.write(json.dumps(to_save, indent=2))
        except IsADirectoryError:
            self.app.notify(f"Could not write to {file_to_save}")
            return
        self.app.notify(f"Wrote config to {file_to_save}")

    @textual.on(textual.widgets.Button.Pressed, "#load")
    def load(self):
        file_to_load = self.context_input.input_path + "/" + self.context_name.value
        try:
            with open(file_to_load, "r") as config_file:
                js = json.load(config_file)
        except (FileNotFoundError, IsADirectoryError):
            self.app.notify(f"Could not find file {file_to_load}")
            return
        except json.JSONDecodeError:
            self.app.notify(f"Error when parsing JSON in {file_to_load}")
            return
        # Delete all other entries
        for ce in self.context_entries:
            ce.remove()
        self.context_entries = []
        self.context_count = 1
        for entry in js:
            self.context_entries.append(ContextEntry(self.context_count))
            self.context_count += 1
            cur_entry = self.context_entries[-1]
            if "archive" in entry:
                cur_entry.archive.input_path = entry["archive"]
            if "extractPaths" in entry:
                for ep in entry["extractPaths"]:
                    cur_entry.extract_paths.add_path(ep)
            if "installPrefix" in entry:
                cur_entry.install_prefix.value = entry["installPrefix"]
            if "containerPrefix" in entry:
                cur_entry.container_prefix.value = entry["containerPrefix"]
        for entry in self.context_entries:
            self.mount(entry, before="#add_context_entry")


# TODO: Rewrite to use ContentSwitcher?
class TUI(textual.app.App):
    """An app for running Surfactant commands"""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("q", "quit", "Quit"),
    ]
    TITLE = "Surfactant TUI"
    CSS_PATH = "../ui-resources/tui.tcss"

    def __init__(self):
        super().__init__()
        self.generate_tab = GenerateTab()
        self.merge_tab = MergeTab()
        self.context_tab = ContextTab()

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Header()
        yield textual.widgets.Footer()
        Tab = textual.widgets.Tab
        with textual.widgets.TabbedContent():
            with textual.widgets.TabPane("Generate"):
                yield self.generate_tab
            with textual.widgets.TabPane("Merge"):
                yield self.merge_tab
            with textual.widgets.TabPane("Context"):
                yield self.context_tab

    def action_toggle_dark(self) -> None:
        """A binding for toggling dark mode"""
        # pylint: disable-next=attribute-defined-outside-init
        self.theme = "textual-dark" if self.theme == "textual-light" else "textual-light"

    def action_quit(self) -> None:
        self.app.exit()


@click.command("tui")
def tui():
    """Run the Surfactant TUI for generating and merging SBOMs."""
    app = TUI()
    app.run()
